/***************************************************************************/
/*                                                                         */
/* IDENTITY:         errmsg                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  12.03.91                */
/*                   CODE             Schlotfeldt  13.03.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  14:58         */
/*                                                                         */
/* FUNCTION:         Generates error messages                              */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'errmsg'                            */
/*                                                                         */
/***************************************************************************/
#include "errmsg.h"

#include "output.h"
#include "strinfo.h"

#include "myalloc.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         errmsg                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Writes error messages via 'output' module to stderr.  */
/*                   Messages (specified by number) may include optional   */
/*                   strings or lists of 'symbols' for output.             */
/*                   Some calls will automatically terminate the program   */
/*                   (fatal errors and help messages).                     */
/*                                                                         */
/* PARAMETERS:       1) Flag: 1 = output of actual input filename and      */
/*                                actual line number before writing the    */
/*                                message                                  */
/*                            0 = write message without filename and line  */
/*                                number                                   */
/*                   2) Message number (0..42)                             */
/*                   3) optional character string to include in the message*/
/*                      instead of '%s' (not used for all messages)        */
/*                   4) optional list of 'symbols' specifying a text, which*/
/*                      will be written after output of the message        */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'output_line' with strings returned from 'get-  */
/*                   name'/'get_line' if 1st parameter flag was set.       */
/*                   Uses 'sprintf' to include the optional string (3rd    */
/*                   parameter if not NULL) in the message before output.  */
/*                   Output is terminated by '\n' or call to 'output_sym'  */
/*                   to generate output for the optional 'symbol' list     */
/*                   (4th parameter if not NULL)                           */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: message 27,28,37,38,39,41,42 will abort the program   */
/*                   message 29                   will exit  the program   */
/*                                                                         */
/***************************************************************************/
void errmsg S4(char,lno,int,mno,char *,exstr,symbol *,exsym)
{
  static struct {                      /* structure containing a message   */
		  char *msg;           /* (as string of characters) and a  */
		  char exit;           /* code specifying wether output of */
		}                      /* the message includes:            */
				       /*   0 = no further actions         */
				       /*   1 = call to 'exit'             */
				       /*   2 = call to 'abort'            */

  err[]=                               /* array of a above structure (all  */
				       /* error messages used by the cpp   */
				       /* project)                         */
  {
    /* 00 */ {"Unterminated string!",                                    0},
    /* 01 */ {"Illegal character constant!",                             0},
    /* 02 */ {"Unterminated character constant!",                        0},
    /* 03 */ {"Unterminated comment!",                                   0},
    /* 04 */ {"Unknown pre-processor command '%s'!",                     0},
    /* 05 */ {"'(' expected!",                                           0},
    /* 06 */ {"')' expected!",                                           0},
    /* 07 */ {"Macro name expected!",                                    0},
    /* 08 */ {"Misplaced '#'!",                                          0},
    /* 09 */ {"Recursive macro '%s'!",                                   0},
    /* 10 */ {"Macro '%s' needs parameter(s)!",                          0},
    /* 11 */ {"Unexpected end of parameter list!",                       0},
    /* 12 */ {"Unmatched ')'!",                                          0},
    /* 13 */ {"Illegal number of parameters to macro '%s'!",             0},
    /* 14 */ {"Error on option '-%s'!",                                  0},
    /* 15 */ {"'%s' : open file failed!",                                0},
    /* 16 */ {"Division by zero!",                                       0},
    /* 17 */ {"':' expected!",                                           0},
    /* 18 */ {"Error directive:",                                        0},
    /* 19 */ {"Parameter name expected!",                                0},
    /* 20 */ {"Parameter names not unique (%s)!",                        0},
    /* 21 */ {"',' or ')' expected!",                                    0},
    /* 22 */ {"Illegal re-definition of '%s'!",                          0},
    /* 23 */ {"End of line expexted!",                                   0},
    /* 24 */ {"'#endif' expected!",                                      0},
    /* 25 */ {"Illegal pragma directive!",                               0},
    /* 26 */ {"<filename> or \"filename\" expected!",                    0},
    /* 27 */ {"Internal error: %s",                                      2},
    /* 28 */ {"Out of memory!",                                          2},
    /* 29 */ {"\
Usage: cpp [options] [file[s]]\n\
       options: -C        Allow nested comments\n\
		-Dxxx     Define macro X\n\
		-Dxxx=yyy Define macro X as Y\n\
		-E0       Disable source line info\n\
		-E1       Enable source line info (default)\n\
		-E2       Debug mode: complete source line info\n\
                -L        don't remove leading blanks\n\
                -R        allow redefinition of macros\n\
		-Ixxx     Include files directory X\n\
		-Uxxx     Undefine macro X\n\
		-inn      Maximum identifier length (>7)\n\
		-i        Unlimited identifier length (default)\n\
		-oxxx     Output file name X\n\
		-h        or\n\
		-?        Print this text\n\
       file[s]: use --xxx for file names starting with -",               1},
    /* 30 */ {"Unmatched '%s'!",                                         0},
    /* 31 */ {"Illegal preprocessor command line!",                      0},
    /* 32 */ {"Recursive #include file '%s'!",                           0},
    /* 33 */ {"Duplicate file '%s' will be ignored!",                    0},
    /* 34 */ {"Illegal line number!",                                    0},
    /* 35 */ {"Illegal constant!",                                       0},
    /* 36 */ {"\"filename\" or end of line expected!",                   0},
    /* 37 */ {"Symbol too complex!",                                     2},
    /* 38 */ {"Macro too complex!",                                      2},
    /* 39 */ {"Line too long!",                                          2},
    /* 40 */ {"Strings can't be merged, string too long!",               0},
    /* 41 */ {"", /* not used */                                         2},
    /* 42 */ {"Can't write output file!",                                2}
  };

  if(lno)                             /* output source line info (if sel.)*/
    output_line(NULL,get_name(),get_line(0),0);

  if(exstr)                            /* generate message text for opt.   */
  {                                    /* string to include...             */
    char *m;

    m=(char *)malloc(strlen(err[mno].msg)+strlen(exstr));
    sprintf(m,err[mno].msg,exstr);
    output_text(NULL,m);               /* ...and pass it to 'output_text'  */
    free(m);
  }
  else                                 /* or use unchanged message for     */
    output_text(NULL,err[mno].msg);    /* output                           */

  if(exsym)                            /* terminate output by writing the  */
    output_sym (NULL,exsym);           /* contents of the opt. 'symbol'    */
  else                                 /* list (the concatenated text) or  */
    output_text(NULL,"\n");            /* simply by passing '\n' to the    */
				       /* 'output' module                  */

  switch(err[mno].exit)                /* process 'exit'/'abort' if 'exit' */
  {                                    /* part of the selected message was */
    case 1: exit(0);                   /* not 0                            */
    case 2: abort();
  }
}
