/***************************************************************************/
/*                                                                         */
/* IDENTITY:         expandm                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  29.03.91                */
/*                   CODE             Schlotfeldt  20.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:03         */
/*                                                                         */
/* FUNCTION:         Detects use of 'macros' in the 'symbol' stream        */
/*                   generated by 'get_lex' function from module 'lexbuf'  */
/*                   and expands those 'macros' to their replacement.      */
/*                   Contains functions get single 'symbols' or a whole    */
/*                   line.                                                 */
/*                   There are 3 different modi of 'macro' replacement     */
/*                   available (see 'next_lex_e')                          */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'next_lex_e', 'next_line',          */
/*                   (+ macros defined in "expandm.h")                     */
/*                   for local use   : 'next_param', 'get_paraml',         */
/*                                     'eval_defined', 'get_lex2'          */
/*                                                                         */
/***************************************************************************/
#include "expandm.h"

#include "lexbuf.h"
#include "macro.h"
#include "srcstack.h"
#include "stringu.h"
#include "lexi.h"

#include "myalloc.h"
#include "errmsg.h"

#define append(lst,tp,len,obj)         /* append:                          */\
{                                      /*   appends to a dynamic allocated */\
  lst=(tp*)realloc                     /*   list 'lst' of length 'len' one */\
	   (lst,++len*sizeof(tp));     /*   element 'obj' of type 'tp'     */\
  if(len<=0) errmsg(1,38,NULL,NULL);                                         \
  lst[len-1]=obj;                                                            \
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_lex2                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Same function as 'get_lex' in module 'lexbuf', but    */
/*                   'symbols' marking the end of a 'symbol' sequence      */
/*                   generated by a 'macro' are skipped and the 'lock'     */
/*                   flag of the attached 'macro' will be cleared.         */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     next 'symbol' from the 'lexbuf' (any 'symbol' except  */
/*                   UNLOCK)                                               */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses 'get_lex'.                                       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Assumes that the next 'symbol' read after UNLOCK      */
/*                   specifies the 'macro' to unlock.                      */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
symbol get_lex2 S0(void)
{
  symbol p;                            /* temporary storage for a 'symbol' */
  macro  m;                            /* temporary storage for a 'macro'  */

  p=get_lex();
  while(p==UNLOCK)                     /* next 'symbol' is UNLOCK ?        */
  {
    if( (m=searchm(get_lex())) )       /* YES: read next 'symbol' specif.  */
      m->lock=0;                       /*      a 'macro' whose 'lock' flag */
    p=get_lex();                       /*      is to be cleared and con-   */
  }                                    /*      tinue with next 'symbol'    */
  return(p);                           /* NO : 'symbol' is the return value*/
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         next_param                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets next parameter for a 'macro' expanded by 'next_  */
/*                   lex_e' as list of 'symbols' from 'get_lex2'           */
/*                                                                         */
/* PARAMETERS:       Flag: 1 = accept NEWLINE 'symbol'                     */
/*                         0 = stop on NEWLINE 'symbol' (like ENDOFFILE)   */
/*                                                                         */
/* RETURN VALUE:     dynamic allocated list of 'symbols' (NULL terminated) */
/*                   or NULL for error                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses 'get_lex2' and 'uget1_lex'. The loop collecting  */
/*                   the 'symbols' exits on error or reaching COMMA or     */
/*                   R_PARENTH (both 'symbols' are ignored if 'pdepth' is  */
/*                   not 0). 'pdepth' counts the unmatched L_PARENTHesis!  */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: calls 'errmsg' on unmatched L_PARENTHesis when        */
/*                   reaching ENDOFFILE or NEWLINE (see parameter)         */
/*                   Return value on error is NULL                         */
/*                                                                         */
/***************************************************************************/
symbol *next_param S1(char,acc_nl)
{
  symbol p,                            /* temporary storage for a 'symbol' */
	 *slist=NULL;                  /* 'symbol' list making the return  */
				       /* value                            */
  size_t sl_len=0,                     /* length of above list             */
	 pdepth=0;                     /* number of unmatched L_PARENTHesis*/

  do p=get_lex2();
  while(p==WHITESP || (p==NEWLINE && acc_nl));
  while((p!=COMMA && p!=R_PARENTH)     /* loop collecting the 'symbols' for*/
	|| pdepth)                     /* the list                         */
  {
    if(p==ENDOFFILE || (p==NEWLINE && !acc_nl))
    {
      if(pdepth)                       /* exit on error for reaching END-  */
      {                                /* OFFILE/NEWLINE with unmatched    */
	free(slist);                   /* L_PARENTHesis                    */
	errmsg( 1,11,NULL,NULL);
	return(NULL);
      }
      break;                           /* normal exit                      */
    }
    if(p==R_PARENTH) pdepth--;
    if(p==L_PARENTH) pdepth++;
    append(slist,symbol,sl_len,p);     /* append current 'symbol' to the   */
    p=get_lex2();                      /* list and continue loop with next */
  }                                    /* 'symbol'                         */
  uget1_lex(p);
  append(slist,symbol,sl_len,NULL);    /* append NULL termination to the   */
  return(slist);                       /* list and return                  */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_paraml                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets complete parameter list as list of lists of      */
/*                   'symbols' from 'get_lex2' calling repeatedly          */
/*                   'next_param'                                          */
/*                                                                         */
/* PARAMETERS:       1) Flag: 1 = accept NEWLINE 'symbol'                  */
/*                            0 = stop on NEWLINE 'symbol' (like ENDOFFILE)*/
/*                   2) 'symbol' (defined as 'macro') whose parameter list */
/*                      is to be parsed (needed for proper error message,  */
/*                      and check for right number of parameters)          */
/*                                                                         */
/* RETURN VALUE:     dynamic allocated list of 'symbol' lists (generated   */
/*                   by 'next_param' (NULL terminated) or NULL for error   */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'next_param' to parse each parameter in the     */
/*                   list. The list should start with optional white space */
/*                   a '(', 0 or more parameters separeted by ',' and a    */
/*                   closing ')'.                                          */
/*                   On error all allocated memory space will be 'freed'   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: returns error (NULL) for missing L/R_PARENTHesis at   */
/*                   the beginning/end of the parsed 'symbol' stream or    */
/*                   if any call to 'next_param' returned NULL (error).    */
/*                   the wrong number of parameters can cause an error too */
/*                                                                         */
/***************************************************************************/
symbol **get_paraml S2(char,acc_nl,symbol,sym)
{
  symbol p,                            /* temporary storage for a 'symbol' */
	 **paraml=NULL;                /* return value (list of lists)     */
  size_t paramc=0;                     /* number of lists in the above list*/
  char   errflg=0;                     /* error flag                       */

  do                                   /* skip optional WHITESPace         */
    p=get_lex2();
  while(p==WHITESP || (p==NEWLINE && acc_nl));
  if(p!=L_PARENTH)                     /* check for L_PARENTH              */
  {
    errmsg( 1,10,sym->text,NULL);
    return(NULL);
  }
  do                                   /* read lists of 'symbols' repre-   */
  {                                    /* senting the parameters...        */
    append(paraml,symbol *,paramc,next_param(acc_nl));
    if(!paraml[paramc-1])
      errflg=1;
    else
      p=get_lex2();
  }
  while(p==COMMA && !errflg);          /* ..til error or reaching the      */
  if(p!=R_PARENTH)                     /* R_PARENTHesis of the list        */
  {
    uget1_lex(p);
    errmsg( 1,11,NULL,NULL);
    errflg=1;
  }
  else                                 /* check for right number of        */
    if(sym->mcr->pc!=paramc)           /* parameters                       */
      if(sym->mcr->pc || paramc!=1 || paraml[0][0])
      {
	errmsg( 1,13,sym->text,NULL);
	errflg=1;
      }
  if(errflg)                           /* on error: 'free' allocated space */
  {                                    /* return value is NULL             */
    while(paramc) free(paraml[--paramc]);
    free(paraml);
    return(NULL);
  }
  append(paraml,symbol *,paramc,NULL); /* append NULL termination          */
  return(paraml);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval_defined                                          */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Parses the parameter list (containing one parameter)  */
/*                   for the ANSI defined function.                        */
/*                   'symbols' to parse are fetched from 'get_lex2'        */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     ONE or ZERO 'symbol' (for parameter of defined-funct. */
/*                   is a macro/is not a macro)                            */
/*                   or NULL for error                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  the DEFINED 'symbol' must have been already read by   */
/*                   the calling function ('next_lex_e')                   */
/*                                                                         */
/* ERROR CONDITIONS: returns error (NULL) for missing L/R_PARENTHesis at   */
/*                   the beginning/end of the parsed 'symbol' stream or    */
/*                   if the parameter between the PARENTHesis was not a    */
/*                   IDENTifier.                                           */
/*                                                                         */
/***************************************************************************/
symbol eval_defined S0(void)
{
# define reterr(n)                     /* reterr:                          */\
  {                                    /*   generate error message number  */\
    errmsg( 1,n,NULL,NULL);            /*   'n' and return with NULL       */\
    return(NULL);                                                            \
  }

  symbol p;                            /* temporary storage for a 'symbol' */
  macro  m;                            /* temporary storage for a 'macro'  */
				       /* here: used as flag               */

  do                                   /* expects: optional WHITESP,       */
    p=get_lex2();
  while(p==WHITESP);
  if(p!=L_PARENTH)   reterr(5)              /*     a '(',                  */
  do
    p=get_lex2();
  while(p==WHITESP);                        /*     optional WHITESP,       */
  if(p->type!=IDENT) reterr(7)              /*     a IDENTifier, which is  */
  m=searchm(p);                             /*     checked to be defined   */
					    /*     as 'macro'              */
  do
    p=get_lex2();
  while(p==WHITESP);                        /*     optional WHITESP,       */
  if(p!=R_PARENTH)   reterr(6)              /*     and a ')'.              */
  return(m?ONE:ZERO);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         next_lex_e                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets next 'symbol' by calling 'get_lex2' from this    */
/*                   module and expands 'macros' to their replace text     */
/*                   (including substitution of their parameters).         */
/*                   'next_lex_e' processes also the stringization operator*/
/*                   (in two different modes) and the defined-function.    */
/*                                                                         */
/* PARAMETERS:       'macro' expand mode:                                  */
/*                     NO_EXP  : selects no 'macro' replacement, just the  */
/*                               stringization operator is processed by    */
/*                               producing 3 'symbols' (QUOTES_, original  */
/*                               'symbol', QUOTES_)                        */
/*                     ALL_EXP : selects complete 'macro' replacement      */
/*                               ('symbols' causing a 'macro' replacement  */
/*                               may be spread over one 1 or more lines)   */
/*                     LINE_EXP: same as ALL_EXP, but string of 'symbols'  */
/*                               causing a 'macro' to be expanded must be  */
/*                               complete before the next line termination.*/
/*                   in ALL_EXP/LINE_EXP mode the stringization operator   */
/*                   produces one STRING 'symbol'!                         */
/*                                                                         */
/* RETURN VALUE:     next 'symbol' for the 'parse' module                  */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation.                             */
/*                   Attention: This function works recursively !!!        */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Recursive 'macros' will be expanded just once and     */
/*                   'errmsg' will be called.                              */
/*                   Misplaced '#' characters ('symbols') will cause an    */
/*                   error too. ('#' must precede a 'symbol' which can be  */
/*                   converted to a not empty string)                      */
/*                                                                         */
/***************************************************************************/
symbol next_lex_e S1(enum exp_mode,expand)
{
  symbol p;                            /* temporary 'symbol' storage       */
  macro  m;                            /* actual 'macro' to replace        */
  symbol sym;                          /* name of 'macro'                  */
  char   *rplc;                        /* actual replace text for the above*/
				       /* 'macro'                          */
  symbol *rplc_s;                      /* same as 'symbol' list            */
  size_t rplc_sl;                      /* length of 'rplc_s' list          */

  char   makestr;                      /* Flag for stringization YES/NO    */
  size_t i;                            /* loop variable                    */

  while(1)
  {
    p=get_lex2();                      /* get next 'symbol' from 'get_lex2'*/

    if(expand==NO_EXP)                 /* NO_EXP mode selected ?           */
    {
      if(p==MAKE_STR)                       /* process stringization ?     */
      {
	p=get_lex2();                       /* yes: generate 3 'symbols' as*/
	uget1_lex(QUOTES_);                 /*      desribed in the funct. */
	uget1_lex(p);                       /*      header                 */
	p=QUOTES_;                          /*      ret.value is the 1st   */
      }
      clear_lex();                          /* no : return just read 'sym- */
      return(p);                            /*      bol'                   */
    }

    if( (makestr=(p==MAKE_STR)) )      /* set stringization flag for other */
      p=get_lex2();                    /* expand modes                     */

    if(p==DEFINED && expand==LINE_EXP) /* check for ANSI-defined-function  */
    {
      p=eval_defined();                     /* evaluate to ICONST 'symbol' */
					    /* ONE or ZERO                 */
      if(!p) goto error;                    /* or use error exit for return*/
    }                                       /* value was NULL              */

    m=searchm(sym=p);                  /* actual 'symbol' is defined as    */
				       /* 'macro' ?                        */
    if(m && m->lock)                   /* check for illegal 'macro'        */
    {                                  /* recursion                        */
      errmsg(1,9,p->text,NULL);
      m=NULL;
    }
    if(!m)
    {                                  /* not a 'macro':                   */
      size_t l;                             /* temporary storage for the   */
					    /* length of the attached text */
					    /* the actual 'symbol'         */
      if(!makestr)                          /* stringization ?             */
      {
	clear_lex();                        /* no : return unchanged actual*/
	return(p);                          /*      'symbol'               */
      }
      if(p==NEWLINE || p==ENDOFFILE)        /* misplaced stringization     */
      {                                     /* operator (at the end of a   */
	errmsg(1, 8,NULL,NULL);             /* line) ?                     */
	clear_lex();
	return(p);
      }
      rplc=(char *)                         /* build replace text starting */
	   malloc((l=strlen(p->text))+2);   /* with quotes and the 'text'  */
      if(l)strcpy(rplc+1,p->text);          /* part of the 'symbol' to     */
      rplc[0]='"';                          /* stringisize                 */
      rplc[l+1]='\0';                       /* (2nd quotes will be appended*/
    }					    /* later)                      */
    else                               /* actual 'symbol' is a 'macro':    */
    {
      char   **parstr;                      /* list of parameter strings   */
					    /* for 'macro' to replace      */
      size_t parn;                          /* number of a parameter (0...)*/

      parstr=(char **)                      /* allocate memory for param.  */
	     malloc((m->pc)*sizeof(char *));/* strings (if any)            */
      if(m->par)                            /* 'macro' needs parameters ?  */
      {
	symbol **paraml;                    /* list of 'symbol' lists re-  */
					    /* presenting parameters       */

	paraml=get_paraml(expand==ALL_EXP,sym);/*parse parameter list      */
	if(!paraml) goto error;             /* error on parameter list ?   */
	clear_lex();                        /* reduce 'lexbuf' all infor-  */
					    /* mation needed has been read */
	for(i=0;i<m->pc;i++)                /* expand recursive 'macros'   */
	{                                   /* contained in the parameters */
	  parstr[i]=NULL;
	  uget1_lex(ENDOFFILE);
	  uget_lexl(paraml[i]);
	  p=next_lex_e(expand);
	  while(p!=ENDOFFILE)               /* and concatenate the 'text'  */
	  {                                 /* fields of the returned      */
	    app_string(&parstr[i],p->text); /* 'symbols'                   */
	    p=next_lex_e(expand);           /* to the strings in the 'par- */
	  }                                 /* str' array                  */
	}
	free(paraml);                       /* 'paraml' is now not longer  */
      }					    /* needed                      */
      rplc=NULL;
      if(makestr)app_string(&rplc,"\"");
      for(i=0;m->rpl[i];i++)
	if(m->rpl[i]->type==IDENT && m->par)
	{
	  parn=0;
	  while(m->par[parn])
	    if(m->par[parn]==m->rpl[i])
	    {
	      app_string(&rplc,parstr[parn]);
	      break;
	    }
	    else
	      parn++;
	  if(!m->par[parn])
	      app_string(&rplc,m->rpl[i]->text);
	}
	else
	  app_string(&rplc,m->rpl[i]->text);
      if(parstr)
      {
	for(i=0;i<m->pc;i++) free(parstr[i]);
	free(parstr);
      }
    }
    if(makestr)app_string(&rplc,"\"");      /* add 2nd quotes to replace   */
					    /* text if 'makestr' flag was  */
					    /* set                         */
    app_string(&rplc,ENDOFFILE->text);      /* mark end of replace string  */
    push_macro(rplc);                       /* call 'push_macro'           */
    rplc_s =NULL;
    rplc_sl=0;
    p=next_lex(0);
    while(p!=ENDOFFILE)                     /* collect 'symbols' resulting */
    {                                       /* from the actual 'macro' ex- */
      append(rplc_s,symbol,rplc_sl,p);      /* pansion to 'rplc_s'         */
      p=next_lex(0);
    }
    append(rplc_s,symbol,rplc_sl,NULL);     /* terminate 'rplc_s' list     */
    if(m)
    {                                       /* activate lock mechanism if  */
      uget1_lex(sym);                       /* replace text was created by */
      uget1_lex(UNLOCK);                    /* a macro (unlock will be done*/
      m->lock=1;                            /* by 'get_lex2'!)             */
    }
    uget_lexl(rplc_s);                      /* put 'rplc_s' list to the    */
    clear_lex();                            /* 'lexbuf' and continue in the*/
  }                                         /* never ending loop...        */

  error:                               /* error exit:                      */
  uget_all_lex();                      /*   write back all 'symbols' read  */
  p=get_lex2();                        /*   since last call to 'clear_lex' */
  return(p);			       /*   or 'uget_all_lex', get the (un-*/
				       /*   changed) first one and return  */
				       /*   with the first 'symbol', whose */
}				       /*   further interpretation caused  */
				       /*   the error                      */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         next_line                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets next line from the source text using the         */
/*                   selected expand mode as string of 'symbols'           */
/*                                                                         */
/* PARAMETERS:       'macro' expand mode (see 'next_lex_e' function)       */
/*                                                                         */
/* RETURN VALUE:     NULL terminated list of 'symbols' representing the    */
/*                   whole next source line (excl. 'symbol' causing the    */
/*                   line termination; it's read but not returned)         */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'next_lex_e'                                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' if there are more 'symbols' to read    */
/*                   than the maximum value for the machine dependend type */
/*                   'size_t'/'sizeof(symbol)'!                            */
/*                                                                         */
/***************************************************************************/
symbol *next_line S1(enum exp_mode,e)
{

# define BLKSIZE 128                   /* minimum size for memory blocks   */
				       /* allocated by 'next_line'         */
				       /* high values minimize calls to    */
				       /* 'realloc'                        */
				       /* should be a power of 2 !!!       */

  symbol p,                            /* actual 'symbol' fetched by 'next-*/
				       /* lex_e'                           */
	 *line=NULL;                   /* list of 'symbols' (return value) */
  size_t length=0;                     /* length of the above list         */

  do
  {
    p=next_lex_e(e);                   /* get next 'symbol'                */
    if(!(length%BLKSIZE))              /* 'realloc' list (if needed)       */
    {
      size_t new;

      new=(length/BLKSIZE+1)*BLKSIZE*sizeof(symbol);
      if(new<=0) errmsg(1,39,NULL,NULL);
      line=(symbol *)realloc(line,new);
    }
    if(endofline(p))p=NULL;            /* NULL termination on end of line  */
    line[length++]=p;                  /* append 'symbol' to list          */
  }
  while(p);                            /* and repeat til NULL termination  */
  return(line);
}
