/***************************************************************************/
/*                                                                         */
/* IDENTITY:         fetch                                                 */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  16.03.91                */
/*                   CODE             Schlotfeldt  13.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:08         */
/*                                                                         */
/* FUNCTION:         Filter for illegal control characters, trigraph       */
/*		     translation, line counting and escape sequences.      */
/*		     Input characters are defined by 'stream' object.      */
/*		     'fetch' should be used instead of 'get_char' (see     */
/*                   module 'input').                                      */
/*                                                                         */
/* DATA/TABLES:      Contains static data (states of filter steps)         */
/*                                                                         */
/* PROCEDURES:       for external use: 'fetch'                             */
/*		     for local use   : 'fetch1', 'fetch2'                  */
/*                                                                         */
/***************************************************************************/
#include "fetch.h"

#include "input.h"

#include "myctype.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         fetch1                                                */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         First filter step (translates trigraphs)              */
/*		     Characters are read from 'stream' via 'get_char'.     */
/*                                                                         */
/* PARAMETERS:       'stream' to use for input, or NULL for initialization */
/*                   of 'fetch1' and 'input' module                        */
/*                                                                         */
/* RETURN VALUE:     Translated character and modified 'stream' object     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'fetch1' knows 6 states (0..5) starting with 3:       */
/*                   0 : 0 '?' are buffered, actual character is the       */
/*                       return value.                               ->3   */
/*                   1 : 1 '?' is buffered, 1 character is stored.         */
/*                       The buffered '?' belongs not to a trigraph and is */
/*                       the return value.	 	             ->0   */
/*                   2 : Same as above: 2 '?' are buffered representing    */
/*                       (together with the stored character) no legal     */
/*			 trigraph.                                   ->1   */
/*                   3 : Start: 0 '?' buffered, no actual character is     */
/*                       stored. Next input character is the return value, */
/*                       if not a '?'. Else:                         ->4   */
/*                   4 : 1 '?' is buffered no actual character is stored.  */
/*                       Next input character is stored.                   */
/*                       If not a '?':                               ->1   */
/*                               else:                               ->5   */
/*                   5 : 2 '?' are buffered no actual character is stored. */
/*                       Next input character is stored.                   */
/*                       If legal character for a trigraph: return trans-  */
/*                       lated trigraph character.                         */
/*                       Else: return '?' if character was a '?' and ->5   */
/*                             else                                  ->2   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'fetch1' must be initialized before the first proper  */
/*                   call with a valid 'stream'.                           */
/*                   Further restrictions see 'get_char' in module 'input'.*/
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char fetch1 S1(stream *,f)
{
  static char *trigr1="()<>/!'-=",     /* list of '??x' characters         */
	      *trigr2="[]{}\\|^~#",    /* list of corresponding characters */
	      *p;                      /* pointer to position in 1st list  */
  static char c,                       /* actual (stored) character        */
	      st;                      /* actual state of fetch1           */

  if(!f)
  {
    st=3;
    init_input();
  }
  else
    while(1) switch(st)
    {
      case 0 : st=3;
	       return(c);
      case 1 :
      case 2 : st--;
	       return('?');
      case 3 : c=get_char(f);
	       if(c!='?') return(c);
      case 4 : c=get_char(f);
	       if(c!='?')
	       {
		 st=1;
		 break;
	       }
	       st=5;
      case 5 : c=get_char(f);
	       p=strchr(trigr1,c);
	       if(!p || !c)
	       {
		 if(c=='?') return('?');
		 st=2;
		 break;
	       }
	       st=3;
	       return(trigr2[p-trigr1]);
    }
  return(0);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         fetch2                                                */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Second filter step (converts illegal control          */
/*                   characters and processes line counting for the given  */
/*                   'stream', which is uses for input via 'fetch1'.       */
/*                                                                         */
/* PARAMETERS:       'stream' to use for input, or NULL for initialization */
/*                   of 'fetch2', 'fetch1' and 'input' module              */
/*                                                                         */
/* RETURN VALUE:     Translated character and modified 'stream' object     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'fetch2' stores the last returned character (EOF on   */
/*		     initialization). Line counting is done on every first */
/*                   character on a line (last character was '\n' or EOF). */
/*                   Other control characters than '\n', '\f') or EOF are  */
/*                   converted to SPC.                                     */
/*                   The first EOF on a just closed 'stream' is converted  */
/*                   to an additional '\n' to simplify recognition of line */
/*                   termination.                                          */
/*                   '\f' will be converted to '\n' (no line counting)     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'fetch2' must be initialized before the first proper  */
/*                   call with a valid 'stream'.                           */
/*                   Further restrictions see 'get_char' in module 'input'.*/
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char fetch2 S1(stream *,f)
{
  static char   c;                     /* actual (stored) character        */
  auto   stream f2;                    /* temp. variable to see if 'f' is  */
				       /* just now closed                  */

  if(!f)
  {
    c=(char)EOF;
    fetch1(NULL);
  }
  else
  {
    if(c=='\n' || c==(char)EOF) f->line++;
    f2= *f;
    c=fetch1(f);
    if(c!='\n' && c!='\f' && iscntrl(c)) c=' ';
    if(f2.fp && c==(char)EOF) c='\n';
  }
  return((c=='\f')?'\n':c);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         fetch                                                 */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Last filter step (processes escape sequences)         */
/*		     Characters are read from 'stream' via 'fetch2'.       */
/*                                                                         */
/* PARAMETERS:       'stream' to use for input, or NULL for initialization */
/*                   of 'fetch', 'fetch2', 'fetch1' (= module 'fetch') and */
/*                   'input' module                                        */
/*                                                                         */
/* RETURN VALUE:     Translated character and modified 'stream' object     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'fetch' knows 4 states (0..3) starting with 2:        */
/*                   0 : No '\\' is buffered, actual character is the      */
/*                       return value.                               ->2   */
/*                   1 : One '\\' is buffered, 1 character is stored.      */
/*                       The buffered '\\' belongs not to a known escape   */
/*                       sequence and is the return value.	     ->0   */
/*                   2 : Start: No '\\' is buffered, no actual character   */
/*                       is stored. Next input character is the return     */
/*			 value if not a '\\'. Else                   ->3   */
/*                   3 : Return single (or no) character for a known esc   */
/*                       sequence and fall back to                   ->2   */
/*                       If unknown:                                 ->1   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'fetch' must be initialized before the first proper   */
/*                   call with a valid 'stream'.                           */
/*                   Further restrictions see 'get_char' in module 'input'.*/
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char fetch S1(stream *,f)
{
  static char c,                       /* actual (stored) character        */
	      st;                      /* actual state of fetch            */

  if(!f)
  {
    st=2;
    fetch2(NULL);
  }
  else
    while(1) switch(st)
    {
      case 0 : st=2;
	       return(c);
      case 1 : st=0;
	       return(ESC);
      case 2 : c=fetch2(f);
	       if(c!='\\') return(c);
      case 3 : c=fetch2(f);
	       st=2;
	       switch(c)
	       {
		 case '\n': break;          /* skip continued line         */
		 case '\\': return('\\');
		 case '\"': return(QUOTES); /* QUOTES, APOSTR and HASHM are*/
		 case '\'': return(APOSTR); /* special characters defined  */
		 case '#' : return(HASHM);  /* in 'types.h'                */
		 case '?' : return('?');
		 default  : st=1;
	       }
    }
  return(0);
}
