/***************************************************************************/
/*                                                                         */
/* IDENTITY:         lexi                                                  */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  18.03.91                */
/*                   CODE             Schlotfeldt  14.04.91                */
/*                                                                         */
/* FUNCTION:         Performs a lexical analysis on the stream of          */
/*                   characters read from 'input' function of module       */
/*		     'src_stack'. Returned 'symbols' are created by        */
/*                   module 'symbol'.                                      */
/*                                                                         */
/* ------------------------------------------------------------------------

   %W% %D%

   Change History:
   ===============
   Apr-25-92:	OTHER --> PPCOTHER
   Mar-22-92:	'yytext' renamed to 'src_yytext' to avoid naming clashes with
   .		'lex' generated modules

   ------------------------------------------------------------------------*/
/*                                                                         */
/* DATA/TABLES:      Contains static data (states of 'next_lex' function)  */
/*                                                                         */
/* PROCEDURES:       for external use: 'next_lex', 'init_lexi'             */
/*                                                                         */
/***************************************************************************/
#include "lexi.h"

#include "srcstack.h"
#include "symbol.h"
#include "errmsg.h"
#include "stringu.h"

#include "myalloc.h"
#include "myctype.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         next_lex                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets next lexem from the buffered input stream using  */
/*                   'input', 'unput', 'unput_all' and 'reduce' function   */
/*                   of module 'src_stack'.                                */
/*                                                                         */
/* PARAMETERS:       Flag for signaling initialization (1), initialization */
/*                   for recognition of nested comments (2) or normal      */
/*                   + (4) for remove leading blanks flag                  */
/*                   operation (0)                                         */
/*                                                                         */
/* RETURN VALUE:     Next lexem ('symbol') as created by 'symbol' module   */
/*                   or NULL on initialization call.                       */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'next_lex' knews 19 different 'states' (starting with */
/*                   0) while analyzing one lexem ('symbol') described in  */
/*                   the external documentation.                           */
/*                   Further there are 4 static flags described below and  */
/*                   one additional flag for nested-comments-on/off.       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Don't use this routine in a not initialized state!    */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
symbol next_lex S1(char,init)
{
  static char    merge,                /* flag for merge/no merge of white */
				       /* space (set by '##')              */
		 first,                /* flag for first non whitespace    */
				       /* character on a line              */
		 empty,                /* flag for "last line was empty"   */
		 incl_l;               /* flag for "actual line belongs to */
				       /* #include-preprocessor-line" used */
				       /* for recognition of <...> strings */
  static char    nestc;                /* nested-comments-on(1)/off(0) ?   */
  static char    removeleadingblanks;  /* flag for remove leading blanks   */
  auto   int     st=0;                 /* internal state (0..17)           */
  auto   char    c;                    /* actual character read via 'input'*/
  auto   size_t  depth=0;              /* depth of comment nesting         */

# define gotost(s)                     /* gotost:                          */\
  {                                    /*   breaks current switch statement*/\
    st=s;                              /*   and continues with the next    */\
    break;                             /*   character at state 's'         */\
  }
# define ignore                        /* ignore:                          */\
  {                                    /*   reduces all characters since   */\
    reduce();                          /*   last 'reduce' call to nothing  */\
    gotost(0)                          /*   and goes to state 0 (start)    */\
  }

# define retlex_(t,u,m,f)              /* retlex_:                         */\
  {                                    /*   returns a trivial 'symbol' 't' */\
    if(u) unput(u);                    /*   after writing back 'u' chars   */\
    reduce();                          /*   and reducing the remaining ones*/\
    merge=m;                           /*   'merge' flag is set to 'm' and */\
    if(!(first=f)) empty=0;            /*   'first' to 'f'. 'empty' flag is*/\
    return(t);                         /*   automatical cleared if 'f' was */\
  }                                    /*   not set.                       */

# define retlexi(t,u)                  /* retlexi:                         */\
  {                                    /*   returns non-trivial 'symbol' of*/\
    if(u) unput(u);                    /*   type 't' which is build by 'mk-*/\
    merge=first=empty=0;               /*   symbol' from all characters    */\
    return(mk_symbol(src_yytext(),t,0l));  /*   read since last 'reduce'.      */\
  }                                    /*   before operation starts there  */
				       /*   be 'u' characters written back */
				       /*   to their source. 'merge', 'emp-*/
				       /*   ty' and 'first' are cleared.   */

  if(init)
  {                                    /* initialization of static flags   */
    merge=first=empty=1;
    incl_l=0;
    removeleadingblanks=init>>2;       /* extract leading blanks flag      */
    init &= 3;
    nestc=(init>1);
    return(NULL);
  }

  while(1)                             /* loop for reading characters from */
  {                                    /* 'input'. loop will be broken by  */
    c=input();                         /* 'retlex_' or 'retlexi'!          */
    switch(st)
    {
      case 0 : switch(c)
	       {
		 case '\n': if(empty)        ignore /* skip double empty   */
						    /* lines               */
			    empty=first;
			    incl_l=0;
					     retlex_(NEWLINE  ,0,1,1)
		 case (char)EOF :                  retlex_(ENDOFFILE,0,0,0)
		 case '~' :		     retlex_(COMPLM   ,0,0,0)
		 case '?' :		     retlex_(QUESTM   ,0,0,0)
		 case '(' :		     retlex_(L_PARENTH,0,0,0)
		 case ')' :		     retlex_(R_PARENTH,0,0,0)
		 case ',' :		     retlex_(COMMA    ,0,0,0)
		 case ':' :		     retlex_(COLON    ,0,0,0)
		 case '%' :                  retlex_(MODULO   ,0,0,0)
		 case '^' :                  retlex_(BIN_XOR  ,0,0,0)
		 case '*' :                  retlex_(MULTIPL  ,0,0,0)
		 case '!' : if(input()=='=') retlex_(NOT_EQUAL,0,0,0)
					else retlex_(NOT      ,1,0,0)
		 case '&' : if(input()=='&') retlex_(LOG_AND  ,0,0,0)
					else retlex_(BIN_AND  ,1,0,0)
		 case '|' : if(input()=='|') retlex_(LOG_OR   ,0,0,0)
					else retlex_(BIN_OR   ,1,0,0)
		 case '+' : if(input()=='+') retlex_(INCREMENT,0,0,0)
					else retlex_(PLUS     ,1,0,0)
		 case '-' : if(input()=='-') retlex_(DECREMENT,0,0,0)
					else retlex_(MINUS    ,1,0,0)
		 case '=' : if(input()=='=') retlex_(EQUAL    ,0,0,0)
					else retlexi(PPCOTHER ,1)
		 case '#' : merge=1;
			    if(input()=='#') gotost(14)
			    unput(1);
			    if(!first)       retlex_(MAKE_STR ,0,1,0)
					else gotost(15)
		 case '/' : if(input()=='*') gotost(12)
					else retlex_(SLASH    ,1,0,0)
		 case '<' : if(incl_l)       gotost(17)
			    switch(input())
			    {
			      case '<' :     retlex_(L_SHIFT  ,0,0,0)
			      case '=' :     retlex_(LSOREQUAL,0,0,0)
			      default  :     retlex_(LESS     ,1,0,0)
			    }
		 case '>' : switch(input())
			    {
			      case '>' :     retlex_(R_SHIFT  ,0,0,0)
			      case '=' :     retlex_(GROREQUAL,0,0,0)
			      default  :     retlex_(GREATER  ,1,0,0)
			    }
		 case ' ' : if (! removeleadingblanks && first)
                               gotost(18)
                            else
                               gotost(14)
		 case 'L' : switch(c=input())
			    {
			      case '"' : gotost(9)  /* w_char type string  */
			      case '\'': gotost(10) /* same for character  */
			      default  : unput(1);
					 gotost(1)
			    }
			    break;
		 case '\'': gotost(10)
		 case '"' : gotost(9)
		 case '.' : if(isdigit(input())) gotost(5)
					    else retlexi(PPCOTHER,1)
		 case '0' : if(tolower(input())=='x')
			    {
			      if(isxdigit(input())) gotost(2)
			      unput(1);
			    }
			    unput(1);
		 default  : if(isdigit(c))
			      gotost(4)
			    else
			    if(is1ident(c))
			      gotost(1)
			    else
			      retlexi(PPCOTHER,0)
	       }
	       break;
      case 16:
      case 1 : if(isident(c)) break;
	       if(st==16)
	       {
		 symbol s;                  /* identifier is preprocessor  */
		 char   *l1,*l2;	    /* command (needs leading '#') */

		 unput(1);
		 l1=strdup("#");
		 app_string(&l1,l2=src_yytext());
		 free(l2);
		 incl_l=((s=mk_symbol(l1,PPCMD,0l))==INCLUDE);
		 retlex_(s,0,0,0)
	       }
	       retlexi(IDENT,1)
      case 2 : if(isxdigit(c)) break;
      case 3 : switch(tolower(c))
	       {
		 case 'u' : if(tolower(input())!='l') unput(1);
			    break;
		 case 'l' : if(tolower(input())!='u') unput(1);
			    break;
		 default  : unput(1);
	       }
	       retlexi(ICONST,0)
      case 4 : if(isdigit(c)) break;
	       switch(tolower(c))
	       {
		 case '.' : gotost(5)
		 case 'e' : gotost(7)
		 default  : unput(1);
			    gotost(3)
	       }
	       break;
      case 5 : if(isdigit(c)) break;
	       switch(tolower(c))
	       {
		 case 'e' : gotost(6)
		 case 'l' :
		 case 'f' : retlexi(FCONST,0)
		 default  : retlexi(FCONST,1)
	       }
	       break;
      case 6 :                         /* process exponent on a number     */
      case 7 :                         /* recognized as INT (state 7) or   */
	       if(c=='+' || c=='-')    /* FLOAT (state 6)                  */
	       {
		 if(!isdigit(input())) retlexi((st==6)?FCONST:ICONST,3)
	       }
	       else
	       if(!isdigit(c)) retlexi((st==6)?FCONST:ICONST,2);
	       gotost(8)
      case 8 : if(isdigit(c)) break;
	       if(tolower(c)!='l' && tolower(c)!='f') unput(1);
	       retlexi(FCONST,0)
      case 17:                         /* process string ending with \" (9)*/
      case 9 : switch(c)               /*                         or > (17)*/
	       {
		 case '\n':
		 case (char)EOF : errmsg(1, 0,NULL,NULL);
			    unput_all();
			    c=input();
			    if(c=='L') retlexi(IDENT,0)
			    retlexi(PPCOTHER,0)
		 case '\"': if(st== 9) retlexi(STRING,0)
			    break;
		 case '>' : if(st==17) retlexi(STRING,0)
			    break;
	       }
	       break;
      case 10:                         /* process character constant       */
      case 11: switch(c)
	       {
		 case '\'': if(st==11) retlexi(CCONST,0)
		 case '\n':
		 case (char)EOF : errmsg(1,(c=='\'')? 1: 2,NULL,NULL);
			    unput_all();
			    c=input();
			    if(c=='L') retlexi(IDENT,0)
			    retlexi(PPCOTHER,0)
		 default  : gotost(11)
	       }
	       break;
      case 12:                         /* skip comments / nested comments  */
      case 13: if(!depth)
	       {
		 unput(3);                  /* unput leading characters on */
		 reduce();                  /* reaching one of the states  */
		 c=input();                 /* from another one ('depth' is*/
	       }                            /* 0)                          */
	       switch(c)
	       {
		 case (char)EOF : unput_all();
			    /* removed for generation of comments
                               errmsg(1, 3,NULL,NULL);
                            */
			    input();
			    retlex_(SLASH,0,0,0);
		 case '*' : if(input()=='/') depth--;
					else unput(1);
			    if(!depth)
			    {
			      reduce();
			      gotost(st+2);
			    }
			    break;
		 case '/' : if(nestc || !depth)
			    {
			      if(input()=='*') depth++;
					  else unput(1);
			    }
	       }
	       break;
      case 18: if(c==' ') break;        /* leading blanks are handled as  */
               unput(1);                /* an identifier                  */
               if(c=='\n' || c==EOF)
                 gotost(0)
               retlexi(IDENT,0)
      case 14:                          /* process whitespace (st==15) for */
      case 15: if(c==' ') break;        /* space between '#' and prepro-   */
	       if(c=='/')               /* cessor command identifier       */
	       {
		 if(input()=='*') gotost(st-2)
			     else unput(1);
	       }
	       if(c=='#')
	       {
		 if(input()=='#')
		 {
		   merge=1;                 /* set 'merge' on '##'         */
		   break;
		 }
		 unput(1);
	       }
	       unput(1);
	       if(st==15)                   /* preprocessor command line?  */
	       {
		 reduce();
		 gotost(16)
	       }
	       if(c=='\n' || c==(char)EOF)        /* ignore whitespace at the    */
		 gotost(0)                  /* end of a line. \n or EOF    */
					    /* will be read again by state */
					    /* 0                           */
	       if(!merge)
		 retlex_(WHITESP,0,0,first) /* whitespace doesn't change   */
					    /* 'first' flag!               */
	       else
		 ignore                     /* ignore merged whitespace    */
    }
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_lexi                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initialization of this module including 'src_stack'   */
/*                   and 'symbol' module                                   */
/*                                                                         */
/* PARAMETERS:       Maximum (significant) length of an identifier or 0    */
/*                   for no limitation.                                    */
/*                   Flag for recognition of nested comments (0=off).      */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'init_lexi' calls 'next_lex' with parameter 1 or 2    */
/*                   dependend on flag for comment nesting and calls       */
/*                   after that 'init_srcstack' and 'init_symbols' with    */
/*                   the max. identifier length.                           */
/*                   'init_symbols' has a build-in self control returning  */
/*                   0 for success, 1 for internal error (passed by        */
/*                   this function to 'errmsg' module)                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  This function should be the called before any call    */
/*                   to 'next_lex'                                         */
/*                                                                         */
/* ERROR CONDITIONS: Failure on 'init_symbols' may cause an internal error */
/*                                                                         */
/***************************************************************************/
void init_lexi S3(long,midl,char,nest,char,removeleadingblanks)
{
  next_lex((nest?2:1)+(removeleadingblanks<<2));
  init_srcstack();
  if(init_symbols(midl))
    errmsg(0,27,"init_symbols()",NULL);
}
