/***************************************************************************/
/*                                                                         */
/* IDENTITY:         parser                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  03.04.91                */
/*                   CODE             Schlotfeldt  26.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  09.06.91  19:32         */
/*                                                                         */
/* FUNCTION:         Parses the stream of 'symbols' generated by           */
/*                   'next_lex_e' (module 'expandm').                      */
/*                   This includes removal of preprocessor command lines   */
/*                   and calling the several functions parsing the command */
/*                   lines found for the single commands implemented.      */
/*                   (#IFxx commands will call recursively this 'module')  */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'parse', 'group', 'init_parser'     */
/*		     for local use   : 'ctrl_line', 'parse_line'           */
/*                                                                         */
/***************************************************************************/
#include "parser.h"

#include "expandm.h"
#include "lexbuf.h"
#include "mergestr.h"
#include "macro.h"

#include "define.h"
#include "undef.h"
#include "error.h"
#include "pragma.h"
#include "include.h"
#include "line.h"
#include "ifsect.h"

#include "errmsg.h"
#include "myalloc.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         parse_line                                            */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Parses a line (previous checked not to be a valid     */
/*                   pp-command line).                                     */
/*                                                                         */
/* PARAMETERS:       1) 'stream' to use for output or NULL                 */
/*                   2) expand mode to use for 'next_lex_e' calls          */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   If 1st parameter is NULL 'parse_line' just skips the  */
/*                   line (reading it 'symbol' by 'symbol' via 'skip2eol').*/
/*                   Otherwise the 'symbols' are collected to a list (by   */
/*                   'next_line') and passed to 'mergestr'.                */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  The line should be terminated by ENDOFLINE or         */
/*                   ENDOFFILE.                                            */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void parse_line S2(stream *,outf,enum exp_mode,em)
{
  if(outf)
  {
    symbol *pa;                        /* temporary storage for list of    */
				       /* 'symbols' (representing the line)*/

    pa=next_line(em);                  /* get the whole line in 'em'       */
    merge_strings(outf,pa,-1);         /* mode, pass line to 'mergestr'    */
  }
  else
  {
    symbol p;                          /* temporary storage for a 'symbol' */

    p=next_lex_e(em);                  /* or skip 'symbols' til end of line*/
    skip2eol(p,em)                     /* if 'outf' was NULL               */
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ctrl_line                                             */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Switches to one of the sub-parser parsing a specific  */
/*                   preprocessor command line or generates an error on    */
/*                   an unknown command.                                   */
/*                   #IFxx commands are not handled by this routine.       */
/*                                                                         */
/* PARAMETERS:       1) 'symbol' specifying the preprocessor command       */
/*                      which was already read via 'next_lex_e' function   */
/*                   2) 'stream' to use for output (here: just as flag)    */
/*                      NULL as 2nd parameter will cause this routine to   */
/*                      ignore the command and skip the remaining part of  */
/*                      pp-command line. Any other value performs normal   */
/*                      action.                                            */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses if-statements to switch to the spec. sub-parser. */
/*                   A switch-statement can't be used, cause 'symbols' as  */
/*                   DEFINE, ERROR etc. are not constant, they are just    */
/*                   used as constants after initialization by module      */
/*                   'symbol'.						   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  First parameter should be already checked to repre-   */
/*                   sent a pp-command.                                    */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on unknown pp-commands */
/*                                                                         */
/***************************************************************************/
void ctrl_line S2(symbol,p,stream *,outf)
{
  if(!outf)      { skip2eol(p,NO_EXP)  /* ignore pp-command line if 2nd    */
		   return;           } /* parameter is NULL                */

  if(p==DEFINE ) { ppDEFINE ();return; }
  if(p==ERROR  ) { ppERROR  ();return; }
  if(p==INCLUDE) { ppINCLUDE();return; }
  if(p==LINE   ) { ppLINE   ();return; }
  if(p==PRAGMA ) { ppPRAGMA ();return; }
  if(p==UNDEF  ) { ppUNDEF  ();return; }

  if(p==NULLDIR)                       /* check NULLDIRective not to be    */
  {                                    /* followed by any other 'symbol'   */
    p=next_lex_e(NO_EXP);              /* as NEWLINE or ENDOFFILE          */
    if(!endofline(p))
      errmsg(1,31,NULL,NULL);
    skip2eol(p,NO_EXP)
    return;
  }

  errmsg(1, 4,p->text,NULL);           /* generate error message for un-   */
				       /* known preprocessor command       */
  uget1_lex(p);                        /* write back 1st 'symbol' and      */
  parse_line(outf,LINE_EXP);           /* 'parse_line' in LINE_EXP mode    */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         group                                                 */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Parses a group of source lines ('symbol' stream is    */
/*                   read via 'next_lex_e'), which is terminated by first  */
/*                   occurence of ENDOFFILE or one of the following pp-    */
/*                   commands: #ELSE, #ELIF or #ENDIF                      */
/*                                                                         */
/* PARAMETERS:       1) 'stream' to use for output or NULL if group is to  */
/*                      be skipped (no output, no command processing)      */
/*                   2) Flag: 1 = 'group' is called by 'if_sect'           */
/*                            0 = 'group' is called by 'parse', group      */
/*                                termination with #ELSE, #ELIF or #ENDIF  */
/*                                is not allowed                           */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Reads each first 'symbol' of a line in NO_EXP mode    */
/*                   (see 'expandm' module) and checks wether this 'symbol'*/
/*                   will terminate the group or leads a pp-command line,  */
/*                   which will be processed by 'ctrl_line'                */
/*                   Normal source lines are parsed by 'parse_line', after */
/*                   the supposed PPCMD 'symbol' is written back to the    */
/*                   'lexbuf'.                                             */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Generates an error message if 3rd parameter was 0 and */
/*                   #ELIF, #ENDIF or #ELSE is detected ('symbols' don't   */
/*                   match to any #IF-command read before)                 */
/*                                                                         */
/***************************************************************************/
void group S2(stream *,outf,char,if_call)
{
  symbol p;                            /* temporary storage for 1 'symbol' */

  p=next_lex_e(NO_EXP);                /* get 1st 'symbol' of a line and   */
  while(p!=ENDOFFILE)                  /* repeate the loop til ENDOFFILE   */
  {                                    /* reached or loop is broken by     */
				       /* #ELIF, #ENDIF or #ELSE           */

    if(p->type==PPCMD)                 /* 1st 'symbol' is pp-command ?     */
    {
	if(p==IF     ||                /* check for if-section,...         */
	   p==IFDEF  ||
	   p==IFNDEF ) if_section(p,outf);
      else                             /* ...if-section termination ...    */
	if(p==ELSE   ||
	   p==ELIF   ||
	   p==ENDIF  ) {
			 if(if_call)
			   break;           /* 'symbols' are illegal for   */
			 else               /* 3rd parameter is 0          */
			   errmsg(1,30,p->text,NULL);
		       }
				       /* ...or normal pp-command line     */
				       /* processed by 'ctrl_line'         */
      else             ctrl_line(p,outf);
    }
    else                               /* else: 1st 'symbol' was not a pp- */
    {                                  /* command                          */
      uget1_lex(p);                    /* write back 1st 'symbol' and      */
      parse_line(outf,ALL_EXP);        /* process as normal line           */
    }
    p=next_lex_e(NO_EXP);              /* get next first 'symbol' on the   */
  }                                    /* following line and continue loop */

  if(if_call)uget1_lex(p);             /* write back ENDOFFILE 'symbol'    */
}                                      /* read as 1st 'symbol' if 2nd      */
				       /* parameter (flag) is set.         */
				       /* 'symbol' must be processed too   */
				       /* by the calling 'if_sect' function*/
				       /* (to generate an error!!)         */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         parse                                                 */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Parses and processes the whole stream of 'symbols'    */
/*                   fetchable from 'next_lex_e' til ENDOFFILE is reached. */
/*                   And flushes after that all buffered lines (module     */
/*                   'mergestr').                                          */
/*                                                                         */
/* PARAMETERS:       'stream' to use for output                            */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'group' and after that 'merge_strings' with     */
/*                   an ENDOFFILE line, to force 'merge_strings' to        */
/*                   generate output for all remaining buffered lines.     */
/*                   (output of the ENDOFFILE line by 'output' module      */
/*                   will finally close the FILE connected to the selected */
/*                   output 'stream')                                      */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'parser' module must be previously initialized by a   */
/*                   call to 'init_parser'.                                */
/*                   First parameter shouldn't be NULL for normal operation*/
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void parse S1(stream *,outf)
{
  auto   symbol *end;                  /* storage for a 'symbol' list      */

  group(outf,0);                       /* parse the whole input generated  */
				       /* by 'next_lex_e'                  */
  end=(symbol *)                       /* create 'symbol' list for the last*/
      malloc(sizeof(symbol)*2);        /* line (must be dynamic allocated, */
  end[0]=ENDOFFILE;                    /* cause 'merge_strings' will set it*/
  end[1]=NULL;                         /* 'free')                          */
  merge_strings(outf,end,-1);          /* flush buffered lines (see 'merge-*/
}                                      /* str' module)                     */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_parser                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initialization of all modules which are directly or   */
/*                   indirectly used by the parser                         */
/*                   The 'parser' module itself contains no static data!   */
/*                                                                         */
/* PARAMETERS:       1) Maximum number of significant characters for an    */
/*                      identifier or 0 for no limit                       */
/*                      (used for 'init_symbol' called by 'init_lexi')     */
/*                   2) Flag: 0 = __STDC__ 'macro' will be undefined       */
/*                            1 = __STDC__ 'macro' will be defined as "1"  */
/*                      (used for 'init_macros')                           */
/*                   3) Flag: 0 = disable recognition of nested comments   */
/*                            1 = enable  recognition of nested comments   */
/*                      (used for 'init_lexi' called by 'init_lexbuf')     */
/*                   4) Mode of source line information output, see module */
/*                      'mergestr'                                         */
/*                   5) remove leading blanks flag                         */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'init_lexbuf', 'init_macros'                    */
/*                   and 'merge_strings'                                   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void init_parser S6(long,midl
                   ,char,stdc
                   ,char,nest
                   ,char,lnos
                   ,char,removeleadingblanks
                   ,char,ignoreredef)
{
  init_lexbuf(midl,nest,removeleadingblanks);
  init_macros(stdc,ignoreredef);
  merge_strings(NULL,NULL,lnos);
}


