/***************************************************************************/
/*                                                                         */
/* IDENTITY:         symbol                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  17.03.91                */
/*                   CODE             Schlotfeldt  13.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  18:45         */
/*                                                                         */
/* FUNCTION:         Creates 'symbols' easy to store and compare           */
/*                                                                         */
/* DATA/TABLES:      Contains static data (hidden in function 'mk_symbol') */
/*                   and one global variable which can be seen as constant */
/*                   (after initialization).                               */
/*                   This variable ('fsymbols') is used to define names    */
/*                   for often used trivial 'symbols'.                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'mk_symbol', 'init_symbols'         */
/*                                                                         */
/***************************************************************************/
#include "symbol.h"

#include "myalloc.h"

symbol fsymbols[FSYMB_N];              /* global array for trivial         */
				       /* 'symbols' as +,-,==, etc.        */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         mk_symbol                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Creates a 'symbol' of given type and text which       */
/*                   was reduced to represent the 'symbol'                 */
/*                                                                         */
/* PARAMETERS:       1) reduced text for 'symbol'                          */
/*                   2) type of 'symbol' eg. IDENT or STRING               */
/*		     3) parameter for initialization:                      */
/*                      <0 : initialization for unlimited significant      */
/*                           length for identifiers (type IDENT)           */
/*                      >0 : initialization for maximum identifier length  */
/*                           expressed by the value itself                 */
/*                      =0 : normal operation                              */
/*                                                                         */
/* RETURN VALUE:     Created 'symbol'                                      */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   if third parameter isn't 0 'mk_symbol' performs an    */
/*                   initialization of its own static variables.           */
/*                   After that it shorts the 'symbol' text if type was    */
/*                   IDENT and initialization was made for a maximum of    */
/*                   significant characters for identifiers.               */
/*                   From the remaining text it creates a 'hash' value     */
/*                   which selects together with the type of 'symbol' one  */
/*                   element of the local static 2-dimensional array       */
/*                   'symbols' representing a sorted list (forward chained)*/
/*                   of 'symbols'.                                         */
/*                   If there exists already an entry for the new 'symbol' */
/*                   no new 'symbol' is allocated. The return value is the */
/*                   same as for the old one created before.               */
/*                   Otherwise a new 'symbol' is allocated and sorted to   */
/*                   the selected list.                                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  First call to 'mk_symbol' must use the initialization */
/*                   parameter!                                            */
/*                   Don't use 'mk_symbol' with third parameter !=0 from   */
/*                   outside this module. Call 'init_symbols' to manage    */
/*                   initialization!                                       */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
symbol mk_symbol S3(char *,s,stype,t,long,init)
{
# define mknewsp(p,t,s)                /* mknewsp:                         */\
  {                                    /*   allocates memory for a new     */\
    p=(symbol)                         /*   'symbol' 'p' and sets its type */\
      malloc(sizeof(struct sym_str));  /*   to 't' and its text to 's'     */\
    p->nxt=NULL;                       \
    p->type=t;                         \
    p->text=s;                         \
    p->mcr=NULL;                       \
  }

# define STABSIZE       50             /* number of sorted lists, big      */
				       /* values decrease searching time   */
				       /* for new 'symbols'                */

  static symbol symbols               /* 2-dimesional array, each element */
		   [STABSIZE]          /* represents a sorted list of      */
		   [last_stype];       /* 'symbols'                        */
  static long   ident_len;             /* max.identifier length or 0 if    */
				       /* unlimited                        */

  unsigned hash=0;                     /* 'hash' value for 'symbol' list   */
				       /* select                           */
  int      sc;                         /* temporary variable for result    */
				       /* of 'strcmp'                      */
  size_t   i;                          /* loop variable                    */
  symbol   new,p,p2;

  if(init)
  {
    int st;                            /* initialization of static data    */

    for(st=0;st<(int)last_stype;st++)
      for(i=0;i<STABSIZE;symbols[i++][st]=NULL);
    ident_len=(init>0)?init:0;
  }

  if(t==IDENT && ident_len)            /* checking max. identifier length  */
    if((long)strlen(s) > ident_len) s[ident_len]=0;

  for(i=0;s[i];hash=(hash<<1)^s[i++]); /* calculating the 'hash' value...  */
  hash%=STABSIZE;                      /* ...max. STABSIZE-1               */
  if(!(p=symbols[hash][(int)t]))
  {
    mknewsp(new,t,s);                  /* create new list for this element */
    symbols[hash][(int)t]=new;         /* of 'symbols' if list was empty   */
    return(new);
  }
  if(strcmp(s,p->text)<0)              /* ...else:                         */
  {                                    /* insert new 'symbol' before the   */
    mknewsp(new,t,s);                  /* element of the selected list if  */
    new->nxt=p;                        /* 'symbol' (its text) is less      */
    symbols[hash][(int)t]=new;         /* than the old first 'symbol'      */
    return(new);
  }                                    /* ...else:                         */
  do                                   /* search insert position (behind   */
  {                                    /* element pointed to by 'p2')      */
    if((sc=strcmp(s,p->text))>=0) p2=p;
    if(!sc)
    {                                  /* exit loop and return if 'symbol' */
      free(s);                         /* with same text was found         */
      return(p);
    }
    p=p->nxt;
  }
  while(p && sc>0);
				       /* ...else:                         */
  mknewsp(new,t,s);                    /* insert new 'symbol' behind 'p2'  */
  new->nxt=p2->nxt;
  p2->nxt=new;
  return(new);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_symbols                                          */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initializes 'mk_symbol' and creates some often used   */
/*                   'symbols' initializing the global variable 'fsymbols' */
/*                                                                         */
/* PARAMETERS:       Maximum significant identfier length or 0 for no      */
/*                   limitation                                            */
/*                                                                         */
/* RETURN VALUE:     0 on success, other if failure occured while          */
/*                   initialization!                                       */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Creates four kinds (types) of 'symbols'. There is     */
/*                   for each type a static (constant) array of strings    */
/*                   containing the individual text.                       */
/*                   'init_symbols' uses 'mk_symbol' to create the         */
/*                   'symbols' which will be stored in 'fsymbols'.         */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Don't change the order of the strings in the static   */
/*                   arrays without changing the 'types.h' file, which will*/
/*                   attach a symbol-name (by #define) to each element of  */
/*                   'fsymbols'. These names are used in many other modules*/
/*                   of this project!                                      */
/*                                                                         */
/* ERROR CONDITIONS: 'init_symbols' compares the number of created         */
/*                   'symbols' with the defined size of 'fsymbols'!        */
/*                   The logical value of this compare (!=) is the return  */
/*                   value.                                                */
/*                                                                         */
/***************************************************************************/
int init_symbols S1(long,ident_len)
{
  int i,j;
  static char *op_sep[] = { "/"   ," "   ,"\n"  ,"!"   ,"%"   ,"^"   ,"&"   ,
			    "*"   ,"-"   ,"+"   ,"~"   ,"|"   ,"<"   ,">"   ,
			    "?"   ,"#"   ,"++"  ,"--"  ,"<<"  ,">>"  ,"<="  ,
			    ">="  ,"=="  ,"!="  ,"&&"  ,"||"  ,"("   ,")"   ,
			    ","   ,":"   ,"\377","\""  ,"UNLK",NULL  };
  static char *ppcmd [] = { "#define"  ,"#if"      ,"#ifdef"   ,"#ifndef"  ,
			    "#elif"    ,"#else"    ,"#endif"   ,"#error"   ,
			    "#include" ,"#line"    ,"#pragma"  ,"#undef"   ,
			    "#"        ,NULL       };
  static char *ident [] = { "defined"  ,"__STDC__" ,"__DATE__" ,"__TIME__" ,
			    "__FILE__" ,"__LINE__" ,NULL       };
  static char *iconst[] = { "0"        ,"1"        ,NULL       };

  j=0;
  fsymbols[j++]=mk_symbol(op_sep[0],OP_SEP,ident_len?ident_len:-1l);

  for(i=1;op_sep[i];fsymbols[j++]=mk_symbol(op_sep[i++],OP_SEP,0l));
  for(i=0;ppcmd [i];fsymbols[j++]=mk_symbol(ppcmd [i++],PPCMD ,0l));
  for(i=0;ident [i];fsymbols[j++]=mk_symbol(ident [i++],IDENT ,0l));
  for(i=0;iconst[i];fsymbols[j++]=mk_symbol(iconst[i++],ICONST,0l));

  return(j!=FSYMB_N);
}
