#include "support.h"
#ifdef TCL_CHANNEL_SUPPORT

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <limits.h>
#include "compat.h"

#include <fcntl.h>

#include <errno.h>

#include "systype.h"
#ifdef BSD
#include <sys/file.h>
#include <sys/ioctl.h>
#endif

#ifdef LINUX
#include  <sys/stat.h>
#endif
#include <termio.h>
#include <sys/ioctl.h>
#include <signal.h>
#include "proto.h"
#include "pty.h"

void pipe_write(int pipefd[2],char *s,int nleft)
{
	int nwritten;
	
	while (nleft > 0) {
		nwritten = write(pipefd[1],s,nleft);
		if (nwritten <= 0) { 
			close(pipefd[1]);
			exit(1);
		}
		nleft -= nwritten;
		s += nwritten;
	}		
}

void pipe_tell(int pipefd[2],char *s)
{
	pipe_write(pipefd,s,strlen(s));
}

void pipe_tellexit(int pipefd[2],char *s)
{
	pipe_write(pipefd,s,strlen(s)+1);
	close(pipefd[1]);
	exit(1);
}

int pipe_wait(int pipefd[2],char *s)
{
	while (read(pipefd[0],s,1) > 0) 
		if (!*s++) {
			close(pipefd[0]);
			return 1;
		}
	close(pipefd[0]);
	return -1;
}

void sig_child()
{
#if defined(BSD) 
	union wait status;
	
	while (wait3(&status, WNOHANG, (struct rusage *)0) > 0);
#endif
}

int Tcl_daemonCmd(ClientData clientData,Tcl_Interp * interp,int argc,char *argv[])
{
    register int childpid,fd;
    char pty_name[PATH_MAX];
    struct termios org_termios;
    struct winsize org_winsize;
    int master_fd, slave_fd;
    int interactive;
    int pipefd[2];
	
    if (argc > 3) {
	interp->result = "wrong # args";
	return TCL_ERROR;
    }
	
    if (pipe(pipefd) < 0) {
	interp->result = "unable creating needed pipe";
	return TCL_ERROR;
    }
	
    errno = 0;

    if ( (childpid = fork() ) < 0) {
	interp->result = "can't fork daemon process";
	return TCL_ERROR;
    } else if (childpid > 0) {
	sprintf(interp->result,"%d",childpid);
	close(pipefd[1]);
	if (pipe_wait(pipefd,interp->result) < 0) {
	    kill(childpid,SIGKILL);
	    interp->result = "pipe communication error";
	    return TCL_ERROR;
	} 
	if (!strlen(interp->result)) {
	    sprintf(interp->result,"%d",childpid);
	    return TCL_OK;
	} else
	    return TCL_ERROR;
    } else 
	close(pipefd[0]);
    
    if ( (interactive = isatty(STDIN_FILENO)) != 0) {
	if ( (tcgetattr(STDIN_FILENO, &org_termios) < 0) ||
	     (ioctl(STDIN_FILENO, TIOCGWINSZ, 
		    (char*)&org_winsize) < 0) ) {
	    pipe_tell(pipefd,"can't read terminal attributes, ");
	    pipe_tellexit(pipefd,strerror(errno));
	}
    }
    
#ifdef SIGTTOU
    signal(SIGTTOU,SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN,SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP,SIG_IGN);
#endif
    signal(SIGHUP,SIG_IGN);
	  
    if (setpgid(getpid(), getpid() ) == -1) {
	pipe_tell(pipefd,"can't change process group, ");
	pipe_tellexit(pipefd,strerror(errno));
    }
    
    if ( (master_fd = pty_master(pty_name)) < 0) {
	pipe_tell(pipefd,"can't create master pty, ");
	pipe_tellexit(pipefd,strerror(errno));
	
    }

#ifndef LINUX
    /* LINUX allways returns -1, at least Kernel 2.0.13 */
    if (setsid() < 0) {
	pipe_tell(pipefd,"can't create new session, ");
	pipe_tellexit(pipefd,strerror(errno));
    }
#endif	
    for (fd = 0; 
#ifdef BSD
	 fd < getdtablesize(); 
#else /* SYS5 */
	 fd < NOFILE; 
#endif
	 fd++)
	if (fd != master_fd && fd != slave_fd && 
	    fd != pipefd[1]) close(fd); 	
    errno = 0;
		
    if ( (slave_fd = pty_slave(master_fd,pty_name)) < 0) {
	pipe_tell(pipefd,"can't create slave pty, ");
	pipe_tellexit(pipefd,strerror(errno));
    }
		
#if defined(TIOCSCTTY) && !defined(CIBAUD)
    if (ioctl(slave_fd, TIOCSCTTY, (char*) 0) < 0) {
	pipe_tell(pipefd,"can't aquire pty control terminal, ");
	pipe_tellexit(pipefd,strerror(errno));
    }
#endif
    
    if (interactive) {
	if ( (tcsetattr(slave_fd, TCSANOW, &org_termios) < 0) ||
	     (ioctl(slave_fd, TIOCSWINSZ, &org_winsize) < 0) ) {
	    pipe_tell(pipefd,"can't set terminal attributes, ");
	    pipe_tellexit(pipefd,strerror(errno));
	}
    }
	
    if ( (dup2(slave_fd, STDIN_FILENO) != STDIN_FILENO) ||
	 (dup2(slave_fd, STDOUT_FILENO) != STDOUT_FILENO) ||
	 (dup2(slave_fd, STDERR_FILENO) != STDERR_FILENO) ) {
	pipe_tell(pipefd,"can't duplicate stdin, stdout, stderr: ");
	pipe_tellexit(pipefd,strerror(errno));
    }
    if (slave_fd > STDERR_FILENO)
	close(slave_fd); 
    
    chdir("/");
    umask(0);

#if defined(BSD)  
    signal(SIGCLD,sig_child);
#else /* SYS5 */
    signal(SIGCLD,SIG_IGN);
#endif /* BSD */
	
    if (argc >= 2) {
	FILE *f;
	
	if ( (f = fdopen(master_fd,"a+")) == NULL) {
	    pipe_tellexit(pipefd,"unable to create master pty stream");
	}
	
	setvbuf(f, NULL, _IONBF, 0);
	Tcl_EnterFile(interp,f,TCL_FILE_READABLE|TCL_FILE_WRITABLE);
	
	if (!Tcl_SetVar(interp,argv[1],interp->result,0)) {
	    fclose(f);
	    pipe_tellexit(pipefd,"unable to set master pty tcl file");
	}
	if (argc == 3) {
		
	    if ( (f=fdopen(slave_fd,"a+")) == NULL) {
		pipe_tellexit(pipefd,"unable to create slave pty stream");
	    }
	    setvbuf(f, NULL, _IONBF, 0);
	    Tcl_EnterFile(interp,f,TCL_FILE_READABLE|TCL_FILE_WRITABLE);
	    if (!Tcl_SetVar(interp,argv[2],interp->result,0)) {
		fclose(f);
		pipe_tellexit(pipefd,"unable to set slave pty tcl file");
	    }
	}
	
    } 
	
    pipe_write(pipefd,"",1);
    close(pipefd[1]);
	  
    interp->result = "0";
    return TCL_OK; 
	
    /* daemon process returns here */
}


#endif /* TCL_CHANNEL_SUPPORT */
