# 
# atexit.tcl
#

# atexit -- command to install a Tcl callback to be invoked when
#	 -- the exit command is evalutated.
#
#   exit -- command to exit process, after all callbacks installed by
#	 -- the atexit command have been invoked.
#

#######################################################################
#
# atexit -- manages atexit callbacks.
#
# The option may be appendUnique, append, prepend, insert, delete,
# clear, set, or list.
# The args depends on the option specified.
#

# The atexit_callbacks list is where we store the 
# installed atexit callbacks.
#

proc atexit {{option list} args} {
  global atexit_callbacks;
  if {[catch {set atexit_callbacks}]} {
    set atexit_callbacks {};
  }

  case $option in {
    set {
      set atexit_callbacks $args;
    }
    appendUnique {
      if {[llength $args] != 1} {
	error {wrong # args : should be "atexit appendUnique callback"};
      }
      set callback [lindex $args 0];
      if {[lsearch $atexit_callbacks $callback] == -1} {
        lappend atexit_callbacks $callback;
      }
    }
    append {
      if {[llength $args] != 1} {
	error {wrong # args : should be "atexit append callback"};
      }
      set callback [lindex $args 0];
      lappend atexit_callbacks $callback;
    }
    prepend {
      if {[llength $args] != 1} {
	error {wrong # args : should be "atexit prepend callback"};
      }
      set callback [lindex $args 0];
      set atexit_callbacks "\{$callback\} $atexit_callbacks";
    }
    insert {
      if {[llength $args] != 2} {
	error {wrong # args : should be "atexit insert before callback"};
      }
      set before   [lindex $args 0];
      set callback [lindex $args 1];
      set l {};
      foreach c $atexit_callbacks {
	if {[string compare $before $c] == 0} {
	  lappend l $callback;
	}
	lappend l $c;
      }
      set atexit_callbacks $l;
    }
    delete {
      if {[llength $args] != 1} {
	error {wrong # args : should be "atexit delete callback"};
      }
      set callback [lindex $args 0];
      set l {};
      foreach c $atexit_callbacks {
	if {[string compare $callback $c] != 0} {
	  lappend l $c;
	}
      }
      set atexit_callbacks $l;
    }
    clear {
      if {[llength $args] != 0} {
	error {wrong # args : should be "atexit clear"};
      }
      set atexit_callbacks {};
    }
    list {
    }
    default {
      error {options, append, prepend, insert, delete, clear, set, or list};
    }
  }
  return $atexit_callbacks;
}

#######################################################################
#
# Hide real exit command.
#

rename exit atexit_exit;

#######################################################################
#
# exit -- Wrapper exit command that first invokes all callbacks installed
#      -- by the atexit command before doing real exit.
#

proc exit {{code 0}} {
  global atexit_callbacks;

  while {1} {

    # Every iteration, we rescan atexit_callbacks, in case
    # some callback modifies it.
    #
    if {[catch {set atexit_callbacks} callbacks]} {
      break;
    }
    if {[llength $callbacks] <= 0} {
      break;
    }
    set        callback  [lindex $callbacks 0];
    set atexit_callbacks [lrange $callbacks 1 end];

    catch {uplevel #0 $callback};
  }

  catch {unset atexit_callbacks};
  catch {atexit_exit $code};
}

