% $Id: is_wellformed.pl,v 1.13 1995/08/26 10:11:01 vdm Exp $

% wellformed predicates ------------------------------------------------------------------

is_wf_specification([spec:_ | Defs]) :-
	is_wf_types(Defs),
	is_wf_rest(Defs).

is_wf_rest(_).

% for types ------------------------------------------------------------------------------

is_wf_types(Defs) :-
						% collect all type expressions
						% and type equations
	all_types(Defs, TypeExprList, TypeEquationList),

						% check used type names
	all_type_names_defined(TypeExprList, TypeEquationList),

						% check double definitions
	no_type_defined_twice(TypeEquationList),

						% check for cyclic defs
	no_cyclic_typedefs(TypeEquationList),

						% check for disjoint union types
	disjoint_union_types(TypeExprList, TypeEquationList),

						% check for disjoint selectors
						% in records
	all_selectors_in_record_type_differ(TypeExprList).


% ----------------------------------------------------------------------------------------

						% all TypeExprs are searched for type identifiers
						% and these are looked up in the type definition list

all_type_names_defined(TypeExprList, TypeEquationList) :-
	traverse_args(TypeExprList, is_type_identifier, exist_typedef(TypeEquationList)).

						% the type expressions of interest
is_type_identifier(['type identifier':_ | _]).
is_type_identifier(['generictype':_ | _]).

						% the check
exist_typedef(['type identifier':Pos, TypeId], TypeDefs) :-
	(
	  member([typedef:_,TypeId | _], TypeDefs)
		-> true
	;
	  message(Pos,['type identifier "', TypeId, '" not defined'])
	).

exist_typedef(['generictype':_, [typevars:_ | TypeVars], TypeExpr], TypeDefs) :-
	generic_types(TypeVars, TypeDefs1),
	append(TypeDefs1, TypeDefs, TypeDefs2),
	all_type_names_defined([TypeExpr], TypeDefs2).

generic_types([], []).

generic_types([ [typevar:_, [identifier:Pos, Name]] | Rest ],
	      [ [typedef:Pos, Name, [any:Pos]] | Rest1 ]) :-
	generic_types(Rest, Rest1).

% ----------------------------------------------------------------------------------------

no_type_defined_twice(TypeDefs) :-
	traverse_list(TypeDefs, type_names_different).

type_names_different([typedef:Pos1, TypeId1 | _ ]
                    ,[typedef:Pos2, TypeId2 | _ ]) :-
	(
	  TypeId1 = TypeId2
		-> message(Pos1, ['type identifier "',TypeId1, '" defined twice']),
		   message(Pos2, ['2. definition'])
	;
	  true
	).

% ----------------------------------------------------------------------------------------

no_cyclic_typedefs(TypeDefs) :-
	no_cyclic_typedefs(TypeDefs, TypeDefs).

no_cyclic_typedefs([], _).
no_cyclic_typedefs([[typedef:_Pos, TypeId, TypeExpr | _TypeInv] | Rest], TypeDefs) :-
	no_cyclic_typedefs([TypeId], TypeExpr, TypeDefs),
	no_cyclic_typedefs(Rest, TypeDefs).
	
no_cyclic_typedefs(TypeIds, TypeExpr, TypeDefs) :-
	traverse_tree(TypeExpr,
		is_cyclic_type_of_interest,
		check_cycle(TypeIds, TypeDefs)).


%	all other type expressions terminate cycle test
%	type identifier expressions repeat search

is_cyclic_type_of_interest([Op:_ | _]) :-
	not(cyclic_type(Op)).

check_cycle([Op:Pos | Args], TypeIds, TypeDefs) :-
	(
	  Op = 'type identifier'
		->  Args = [TypeId],
		    ( member(TypeId, TypeIds)
			-> message(Pos, ['cyclic type definition for "', TypeId, '"'])
		    ;
		      member([typedef:_, TypeId, TypeExpr1 | _ ], TypeDefs),
		      no_cyclic_typedefs([TypeId | TypeIds], TypeExpr1, TypeDefs)
			-> true
		    ;
		      true
		    )
	;
	  true
	).

% ----------------------------------------------------------------------------------------

disjoint_union_types(TypeExprList, TypeDefs) :-
	traverse_args(TypeExprList,
		is_union_type,
		disjoint_union_type(TypeDefs)).

is_union_type([uniontype:_ | _]).
is_union_type([optionaltype:_, _]).			% optional is union with nil

disjoint_union_type([uniontype:_ | ComponentList], TypeDefs) :-
	traverse_list(ComponentList, disjoint_types(TypeDefs, []-[])).

disjoint_union_type([optionaltype:Pos, Texpr], TypeDefs) :-
	disjoint_union_type([uniontype:Pos, [nil:Pos], Texpr], TypeDefs).


% error situations -----------------------------------------------------------------------
%
% any indicates an arbitrary type, generated in error situations

disjoint_types([any:_ | _], _, _, _) :-
	!.

disjoint_types(_, [any:_ | _], _, _) :-
	!.


% same type occurs twice -----------------------------------------------------------------

disjoint_types([nil:Pos1], [nil:Pos2], _, _) :-
	!,
	message(Pos1, ['2 optional types in union or optional']),
	message(Pos2, ['2. occurence of type']).

disjoint_types([T1:Pos1|Args1], [T1:Pos2|Args1],_,_) :-
	!,
	message(Pos1, ['same type occurs twice in union type']),
	message(Pos2, ['2. occurence of type']).


% numeric types --------------------------------------------------------------------------

disjoint_types([T1:Pos1 | _], [T2:Pos2 | _],_,_) :-
	numeric_type(T1),
	numeric_type(T2),
	!,
	message(Pos1, ['numeric types in union not disjoint']),
	message(Pos2, ['2. occurence of type']).


% collections of same kind ---------------------------------------------------------------

disjoint_types([T1:Pos1 | _], [T1:Pos2 | _], _, _) :-
	empty_collection_type(T1),
	!,
	message(Pos1, ['collection types in union not disjoint']),
	message(Pos2, ['2. occurence of type']).
	

% records and functions ------------------------------------------------------------------

disjoint_types([T1:Pos1|Args1],
	       [T1:Pos2|Args2],TypeDefs,_) :-
	structured_type(T1),
	!,
	(
	  different_type_lists(Args1,Args2,TypeDefs)
		-> true
	;
	  message(Pos1, ['product types in union not disjoint']),
	  message(Pos2, ['2. occurence of type'])
	).


% unions ---------------------------------------------------------------------------------

disjoint_types([uniontype:_ | Texpr], T2, TypeDefs, L) :-			% 1. arg a union
	!,
	traverse_list([T2 | Texpr], disjoint_types(TypeDefs, L)).

disjoint_types(T1, [uniontype:_, TExpr], TypeDefs, L) :-			% 2. arg a union
	!,
	traverse_list([T1 | TExpr], disjoint_types(TypeDefs, L)).


% optionals ------------------------------------------------------------------------------
%
% optionals are handled as nil | type
% so a 3 element list is built and checked like a variant list
% in a union

disjoint_types([optionaltype:Pos1, TExpr], T2, TypeDefs, L) :-		% 1. arg an optional
	!,
	traverse_list([[nil:Pos1], TExpr, T2], disjoint_types(TypeDefs, L)).

disjoint_types(T1, [optionaltype:Pos2, TExpr], TypeDefs, L) :-		% 2. arg an optional
	!,
	traverse_list([T1, [nil:Pos2], TExpr], disjoint_types(TypeDefs, L)).

% parenthesis around type expressions ----------------------------------------------------

disjoint_types([parenttype:_, T1], T2, TypeDefs, L) :-
	!,
	disjoint_types(T1, T2, TypeDefs, L).

disjoint_types(T1, [parenttype:_, T2], TypeDefs, L) :-
	!,
	disjoint_types(T1, T2, TypeDefs, L).

% type identifier ------------------------------------------------------------------------
%
% the 4. parameter contains a pair of lists
% the lists contain all type identifiers collected during
% type comparison, to prevent infinite loops in cyclic structures

disjoint_types(['type identifier':Pos1, TypeName], T2, TypeDefs, L1-L2) :-
	not(member(TypeName, L1)),
	!,
	(
	  member([typedef:_, TypeName, [Op:_ | Args] | _], TypeDefs),
	  !
		-> disjoint_types([Op:Pos1 | Args], T2, TypeDefs, [TypeName | L1]-L2)
	;
	  true
	).

disjoint_types(T1,T2,TypeDefs,L1-L2) :-
	T2 = ['type identifier':_,TypeName],
	not(member(TypeName,L2)),
	!,
	disjoint_types(T2,T1,TypeDefs,L2-L1).

disjoint_types(['type identifier':Pos1,TypeName1],
	       ['type identifier':Pos2,TypeName2],_,_) :-
	!,
	message(Pos1, ['types "', TypeName1, '" and "',TypeName2, '" in a union have the same structure']),
	message(Pos2, ['definition of "', TypeName2]).

% all things ok --------------------------------------------------------------------------

disjoint_types(_,_,_,_).

% record lists ---------------------------------------------------------------------------

different_type_lists([],[_|_],_).
different_type_lists([_|_],[],_).
different_type_lists([H1|R1],[H2|R2],TypeDefs) :-
	( different_types(H1,H2,TypeDefs)
		-> true
	;
	  different_type_lists(R1,R2,TypeDefs)
	).

% this is too simple !!!
% it must be refined !!!

different_types([T1:_|_], [T2:_|_], _) :-
	T1 \= T2.

% ----------------------------------------------------------------------------------------

all_selectors_in_record_type_differ(TypeDefs) :-
	traverse_args(TypeDefs,
		is_record_type,
		check_selectors).

is_record_type([record:_|_]).

check_selectors([record:_,_Tag | FieldList]) :-
	traverse_list(FieldList,check_field_id).

check_field_id([field:Pos1,Fid1,_],[field:Pos2,Fid2,_]) :-
	(
	  Fid1 = Fid2,
	  Fid1 \= '_'
		-> message(Pos1, ['field selector "',Fid1 ,'" used twice in a record type']),
		   message(Pos2, ['2. occurence of selector'])
	;
	  true
	).

% ----------------------------------------------------------------------------------------

all_types(Defs, TypeExprList, TypeEquations) :-
	all_type_equations(Defs, TypeEquations),
	all_type_expr(Defs, TypeExprList).

all_type_equations(Defs, TypeEquations) :-
	findall(X, (member([types:_ | Y], Defs), member(X,Y)), TypeEquations).

all_type_expr(Defs, TypeExprList) :-
	findall(X, (member(Def, Defs), contains_type_expr(Def, X)), TypeExprList).

contains_type_expr(Tree, Expr) :-
	contains_expr(Tree, contains_a_type_expr, Expr).

contains_a_type_expr([typedef:_, _, TypeExpr, _], TypeExpr).

contains_a_type_expr([patterntypepair:_, TypeExpr | _], TypeExpr).

contains_a_type_expr([idtypepair:_, _, TypeExpr], TypeExpr).

contains_a_type_expr([explicitfct:_, _, ResultType | _], ResultType).

contains_a_type_expr([localdef:_, _, TypeExpr | _], TypeExpr) :-
	TypeExpr \= [ any:_ ].

contains_a_type_expr([typebind:_, TypeExpr | _], TypeExpr).

contains_a_type_expr([value:_, _, TypeExpr | _], TypeExpr) :-
	TypeExpr \= [ any:_ ].

contains_a_type_expr([state:_, RecordType | _], RecordType).

% ----------------------------------------------------------------------------------------