% $Id: transform.pl,v 1.5 1995/06/12 08:48:55 vdm Exp $
% general tree traversal and tre transformation predicates -------------------------------

% ----------------------------------------------------------------------------------------

% traverse_tree(+Tree,+Predicate,+Action) 
%       top down tree traversal through Tree
%       every node is checked, whether Predicate holds for it
%       if that is the case Action is called with this Node as first argument
%       else the children of the tree are processed recursively

traverse_tree(Tree,IsNode,ProcessNode) :-
        Predicate =.. [IsNode,Tree],
        call(Predicate),
        !,
        ProcessNode =.. [Functor|Args],
        Process     =.. [Functor,Tree|Args],
        call(Process).

traverse_tree([_Op|Args],IsNode,ProcessNode) :-
        !,
        traverse_args(Args,IsNode,ProcessNode).

traverse_tree(_,_,_).

traverse_args([],_,_).
traverse_args([Arg|Args],IsNode,ProcessNode) :-
        traverse_tree(Arg, IsNode,ProcessNode),
        traverse_args(Args,IsNode,ProcessNode).


% call a predicate for all pairs of a list -----------------------------------------------
% the predicate and the additional parameters are submitted in arg ProcessPair

traverse_list([],_ProcessPair).

traverse_list([H|T],ProcessPair) :-
	traverse_list(T,H,ProcessPair),
	traverse_list(T,ProcessPair).

traverse_list([],_,_).

traverse_list([H2|R],H1,ProcessPair) :-
	ProcessPair =.. [Functor|Args],
	Process =.. [Functor,H1,H2|Args],
	call(Process),
	traverse_list(R,H1,ProcessPair).

% general transformation strategies ------------------------------------------------------

repeat_transformation(Transform,Expr,Result) :-
	Goal =.. [Transform,Expr,Result1],
	call(Goal),
	!,
	repeat_transformation(Transform,Result1,Result).

repeat_transformation(_,Expr,Expr).

% transformation rules -------------------------------------------------------------------

% associativity

trans_assoc1([Op:Pos|Args],[Op:NewPos|NewArgs]) :-
	assoc_op(Op),
				% search for the associative operator
	append(Args0,[[Op:Pos1|Args11]|Args1],Args),
				% and concatenate the 3 argument lists
	newposition(Pos,Pos1,NewPos),
	append(Args0,Args11,Args01),
	append(Args01,Args1,NewArgs).

trans_assoc(Expr,Result) :-
	repeat_transformation(trans_assoc1,Expr,Result).

% ----------------------------------------------------------------------------------------

trans_left_assoc1([Op:Pos2,Arg1,[Op:Pos1,Arg2,Arg3]],
                  [Op:Pos1,[Op:Pos2,Arg1,Arg2],Arg3]) :-
	left_assoc_op(Op).

trans_left_assoc(Expr,Result) :-
	repeat_transformation(trans_left_assoc1,Expr,Result).

% ----------------------------------------------------------------------------------------

trans_right_assoc1([Op:Pos1,[Op:Pos2,Arg1,Arg2],Arg3],
		   [Op:Pos2,Arg1,[Op:Pos1,Arg2,Arg3]]) :-
	right_assoc_op(Op).

trans_right_assoc(Expr,Result) :-
	repeat_transformation(trans_right_assoc1,Expr,Result).

% ----------------------------------------------------------------------------------------

trans_assoc_expr(Expr,Result) :-
	Expr = [Op:_|_],
	(
	  assoc_op(Op)
		-> trans_assoc(Expr,Result)
	;
	  left_assoc_op(Op)
		-> trans_left_assoc(Expr,Result)
	;
	  right_assoc_op(Op)
		-> trans_right_assoc(Expr,Result)
	;
	  Expr = Result
	).

% ----------------------------------------------------------------------------------------

% transforms an expression into a product expression: Op = producttype
% used for parameter lists

trans_paramlist(E, E) :-			% already a product
        E = [producttype:_ | _],
        !.
trans_paramlist(E, [producttype:Pos, E]) :-	% a product with 1 component
        E = [_:Pos | _].

% ----------------------------------------------------------------------------------------

% contains_expr(+Tree, +Predicate, -Expr)
% looks for an expression that fullfills Predicate
% and returns this, else all arguments are searched
% for subexpressions for which Predicate holds

contains_expr(Tree, Predicate, Expr) :-
	Predicate =.. [ Functor | Args ],
	Predicate1 =.. [ Functor , Tree , Expr | Args ],
	call(Predicate1).

contains_expr([_Op | Args], Predicate, Expr) :-
	list_contains_expr(Args, Predicate, Expr).

list_contains_expr([Tree | _Rest], Predicate, Expr) :-
	contains_expr(Tree, Predicate, Expr).

list_contains_expr([_Tree | Rest], Predicate, Expr) :-
	list_contains_expr(Rest, Predicate, Expr).

% ----------------------------------------------------------------------------------------
% ----------------------------------------------------------------------------------------

