/***************************************************************************/
/*                                                                         */
/* IDENTITY:         cmdline                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  08.04.91                */
/*                   CODE             Schlotfeldt  03.05.91                */
/*                   VERSION(CURRENT) Uwe                                  */
/*                                                                         */
/* FUNCTION:         Parses the contents of the command line passed to     */
/*		     'main' and initializes all modules of this project in */
/*                   the right order with the default and/or specified     */
/*                   parameters (depending on the flags on the passed      */
/*                   command line).                                        */
/* ------------------------------------------------------------------------

   %W% %D%

   Change History:
   ===============
   Feb-04-93:   new flag removeleadingblanks (default true) -L flag
   Jan-20-93:   midl (maximum id length) default set to unlimited
   Apr-25-92:	'usage' corrected ('=' --> '==')
   Apr-25-92:	OTHER --> PPCOTHER

   ------------------------------------------------------------------------*/
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'parse_cmdl_opt'                    */
/*		     for local use   : 'extr_clp', 'usage'                 */
/*                                                                         */
/***************************************************************************/

#include "cmdline.h"

#include "parser.h"
#include "symbol.h"
#include "macro.h"
#include "input.h"
#include "output.h"
#include "srcstack.h"

#include "errmsg.h"
#include "myalloc.h"
#include "myctype.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         extr_clp                                              */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Extracts from a given command line (same parameters   */
/*                   as to 'main') the last occurence of a parameter marked*/
/*                   with a specified flag ('-' + 1 character)             */
/*                                                                         */
/* PARAMETERS:       1) Character specifying the flag to look for          */
/*                      or '\0' for looking at parameters not preceded by  */
/*                      a flag (input filenames)                           */
/*                   2+3) command line (as passed to 'main')               */
/*                                                                         */
/* RETURN VALUE:     Text of command line parameter (excluding the 2       */
/*                   leading characters for the flag) or NULL if there was */
/*                   no parameter attached to the specified flag.          */
/*                   + modified command line (2nd+3rd parameter to this    */
/*                     function) now reduced by the found parameter        */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Scans the parameter list from back to front. A        */
/*                   parameter is found if the actual command line param.  */
/*                   starts with a '-' (not '--') and the 2nd character    */
/*                   of this cl-parameter is equal to the spec. flag or    */
/*                   it's found if cl-parameter starts with '--' or any    */
/*                   other characters (specifying a filename) and the 1st  */
/*                   parameter to this funct. is '\0'                      */
/*                   The cl-parameter list will then be reduced (removing  */
/*                   of the found parameter).                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  command line consists of parameters with flags (for   */
/*                   example "-Iinclude", where 'I' is the flag and        */
/*                   "include" is the parameter text) and parameters       */
/*                   without flags (for example "--filename" or "filename",*/
/*                   where '--' is used for filenames starting with a      */
/*                   single '-')                                           */
/*                                                                         */
/* ERROR CONDITIONS: Returns NULL if searched parameter type wasn't        */
/*                   found.                                                */
/*                                                                         */
/***************************************************************************/
char *extr_clp S3(char,flag,int *,argc,char **,argv)
{
  int i,j;                             /* loop variables                   */
  char *e=NULL;                        /* result of function               */

  for(i= *argc-1;i>0;i--)              /* scan cl-parameter list           */
  {
    if(argv[i][0]=='-')                /* look for parameters...           */
    {
      if((argv[i][1]==flag && flag)    /* ...with spec. flag               */
      || (argv[i][1]=='-' && !flag))   /* ...without flag (and 'flag'==0)  */
	e=strdup(argv[i]+(flag?2:1));
    }
    else
      if(!flag)                        /* ...without flag (and 'flag'==0), */
	e=strdup(argv[i]);             /*    starting with any other       */
				       /*    character than '-'            */

    if(e)                              /* remove parameter (copied to 'e') */
    {                                  /* from the cl-parameter list if    */
				       /* parameter was found              */
      for(j=i;j<*argc;j++) argv[j]=argv[j+1];
      (*argc)--;
      return(e);                       /* return searched parameter text   */
    }
  }
  return(NULL);                        /* or NULL if parameter wasn't      */
}                                      /* found (after finishing the loop) */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         usage                                                 */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Generates an optional error message an a help text    */
/*                   (printed by 'errmsg' too) explaining the right syntax */
/*                   of the command line.                                  */
/*                                                                         */
/* PARAMETERS:       Character specifying the flag to generate an error    */
/*                   message for ("error on flag -x", where 'x' is the     */
/*                   spec. character)                                      */
/*                   There will be no error message for calling 'usage'    */
/*                   with parameter 0 !                                    */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'errmsg'                                        */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Exits to the calling program (eg. operating system)   */
/*                   in any case!                                          */
/*                                                                         */
/***************************************************************************/
void usage S1(char,opt)
{
  char *err="x";

  if(err[0]==opt)
    errmsg(0,14,err,NULL);
  errmsg(0,29,NULL,NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         parse_cmdl_opt                                        */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See module header                                     */
/*                                                                         */
/* PARAMETERS:       1+2) command line (as passed to 'main')               */
/*                                                                         */
/* RETURN VALUE:     'stream' to use for output or NULL if creating the    */
/*                   'stream' object by 'open_w' failed                    */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Phases of cl-parameter analysis / initialization:     */
/*                   0. look for '-?' or '-h' to output help message       */
/*                   1. look for '-i' flag (opt. followed by 1 to 2 decimal*/
/*                      digits to specify the maximum significant length of*/
/*                      an identifier) and set 'midl' (default is unlimited)      */
/*                   2. look for '-C' flag, setting the 'nest' flag (for   */
/*                      recognition of nested comments, default is 0=OFF)  */
/*                   3. look for '-E' flag followed by a digit specifying  */
/*                      the mode of source line information output         */
/*                      (0=no information, 1=information via #line-format  */
/*                      output (default) or 2=information to each line)    */
/*                   4. Initialization call to the 'parser' (includes      */
/*                      initialization of all other modules) using the     */
/*                      results of step 1, 2 and 3                         */
/*                   5. look for parameters marked be the '-I' flag (spec. */
/*                      include-directories for global file search)        */
/*                   6. look for filenames specifying the source (concate- */
/*                      nation of files)                                   */
/*                   7. look for parameter marked with the '-o' flag (spec.*/
/*                      the name of a file to use for output, attached     */
/*                      'stream' is result of this function)               */
/*                   8. look for parameters with '-D' or '-U' flag (spec.  */
/*                      'macros' to define/undefine)                       */
/*                   steps 5, 6 and 8 will call the affected modules to do */
/*                   further initialization!                               */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Any error (syntax, illegal macro defintions) will     */
/*                   cause a call to 'usage'. If the output file (specified*/
/*                   by '-o' parameter (or default: stdout) can't be       */
/*                   opened, this error is expressed by return value NULL. */
/*                   Any other errors will exit the program (see 'usage')  */
/*                                                                         */
/***************************************************************************/
stream *parse_cmdl_opt S2(int,argc,char **,argv)
{
  long   midl=0;                       /* maximum significant length for   */
				       /* identifiers (default: unlimited)        */
  int	 i;                            /* loop variable                    */
  char   *par,                         /* temporary storage for return     */
				       /* values (from 'extr_clp')         */
	 nest=0,                       /* flag for nested comments (default*/
				       /* is OFF)                          */
	 lnos=1,                       /* switch for source line infor-    */
				       /* mation (0, 1=default or 2)       */
         ignoremacroredefinition=0,    /* switch for redefining macros     */
         removeleadingblanks=1;        /* switch for leading blanks        */
  stream *source,                      /* temporary storage for a input    */
				       /* 'stream'                         */
	 *dest;                        /* result of 'parse_cmdl_opt'       */

  /********** step 0: look for '-?' or '-h' to output help message *********/

  if( (par=extr_clp('?',&argc,argv)) )
    usage(par[0]?'?':0);                    /* any additional param.text is*/
					    /* illegal                     */
  if( (par=extr_clp('h',&argc,argv)) )
    usage(par[0]?'h':0);                    /* any additional param.text is*/
					    /* illegal                     */

  /************** step 1: look for '-i' flag to modify 'midl' **************/

  if( (par=extr_clp('i',&argc,argv)) )      /* '-i' flag used ?                 */
  {                                         /* allow flag to be followed by*/
    if(strlen(par) > (size_t)2) usage('i'); /* max. 2 additional characters*/
    if(par[0])                              /* set 'midl' to a spec.value  */
    {                                       /* (expressed by 1 or 2 decim. */
      if(!isdigit(par[0])) usage('i');      /* digits)                     */
      if(!par[1])
	midl=par[0]-'0';
      else
      if(!isdigit(par[1])) usage('i');
      else
	midl=(par[1]-'0')+(par[0]-'0')*10;
      if(midl<8) usage('i');
    }
    else                                    /* 'midl' to 0 (unlimited      */
      midl=0;                               /* significant length) if par. */
  }                                         /* text was empty              */

  /************** step 2: look for '-C' flag to modify 'nest' **************/

  if( (par=extr_clp('C',&argc,argv)) )      /* '-C' flag used ?                 */
  {
    if(par[0]) usage('C');                  /* any additional param.text is*/
    nest=1;                                 /* illegal, occurence of '-C'  */
					    /* sets 'nest' flag (default 0)*/
  }

  /************** step 3: look for '-E' flag to modify 'lnos' **************/

  if( (par=extr_clp('E',&argc,argv)) )      /* '-E' flag used ?                 */
    switch(par[0])                          /* switch on first character of*/
    {                                       /* parameter text              */
      case '0':
      case '1':
      case '2': lnos=par[0]-'0'; break;     /* set 'lnos' to a spec. value */
      case  0 : lnos=1;          break;     /* set 'lnos' to 1 (default)   */
      default : usage('E');                 /* generate error message for  */
					    /* any other character         */
    }

  /************** step 3.5: look for '-L' flag for leading blanks **********/

  if( (par=extr_clp('L',&argc,argv)) )
  {
    if(par[0]) usage('L');                  /* any additional param.text is*/
    removeleadingblanks=0;                  /* illegal, occurence of '-L'  */
  }

  /************** step 3.6: look for '-R' flag for macro redefinition ******/

  if( (par=extr_clp('R',&argc,argv)) )
  {
    if(par[0]) usage('R');                  /* any additional param.text is*/
    ignoremacroredefinition=1;
  }

  /******** step 4: initialize all modules by calling 'init_parser' ********/

  init_parser(midl,0,nest,lnos,removeleadingblanks,ignoremacroredefinition);

  /********* step 5: look for '-I' flag to add include directories *********/

  while( (par=extr_clp('I',&argc,argv)) )  /* repeat until no further '-I' par.*/
  {                                    /* is found                         */
    if(!par[0]) usage('I');                 /* there must be some addition.*/
					    /* param.text to specify the   */
					    /* directory name              */
    add_incl_dir(par);                      /* add name to the static list */
					    /* in module 'input'           */
  }

  /****** step 6: look for parameters without flags (input filenames) ******/

  par=extr_clp(0,&argc,argv);          /* test if thereis at least one in- */
  if(par)                              /* put filename                     */
    do
    {
      if( (source=open_r(par,1)) )         /* try to open spec. file (create   */
      {				       /* 'stream') in the local directory */
	if(push_stream(source))        /* try to push 'stream' on the 'src-*/
	{                              /* stack', generate error for dupli-*/
	  errmsg(0,33,par,NULL);       /* cate file (will be ignored)      */
	  free(source->name);
	  free(source);
	}
      }
      else
      {
	errmsg(0,15,par,NULL);         /* generate error message for open  */
	usage('?');                    /* file / creating 'stream' failed  */
				       /* and print help text without      */
				       /* additional error message         */
      }
      par=extr_clp(0,&argc,argv);      /* continue with next filename (if  */
    }                                  /* found)                           */
    while(par);
  else
    push_stream(open_r(NULL,1));       /* push 'stream' for stdin to the   */
				       /* 'srcstack' if there was no file- */
				       /* name specified by the command    */
				       /* line                             */

  /******** step 7: look for '-o' flag to create a output 'stream' *********/

  if( (par=extr_clp('o',&argc,argv)) )     /* '-o' flag used ?                 */
  {
    if(!par[0]) usage('o');                 /* any additional param.text is*/
					    /* illegal                     */
    dest=open_w(par);                       /* try to open the spec. file  */
					    /* for output                  */
  }                                         /* (created 'stream' is the    */
  else                                      /* return value of this funct. */

    dest=open_w(NULL);                      /* create 'stream' for stdout, */
					    /* if '-o' flag isn't used     */

  /** step 8: look for remaining parameters (scanning from front to end) ***/

  for(i=1;i<argc;i++)                  /* scan remaining cl-parameter list */
  {
    int    j;                          /* pointer to param.text            */
    symbol mname;                      /* 'macro' name to define/undefine  */
				       /* as 'symbol'                      */
    char   sep;                        /* character behind the extracted   */
				       /* 'macro' name in the param.text   */

    if(argv[i][0]!='-')                /* allowed are only parameters with */
      usage(argv[i][1]);               /* flags ('-D' or '-U', checked be- */
    switch(argv[i][1])                 /* low:)                            */
    {
      case 'D' :                       /* check first character of param.  */
      case 'U' :                       /* text (following the flag) to be  */
				       /* valid for a identifier ('macro'  */
				       /* name                             */
	   if(!is1ident(argv[i][2]))
	     usage(argv[i][1]);
	   j=3;
	   while(isident(argv[i][j]))  /* parse remaining characters, which*/
	     j++;                      /* belong to the identifier         */
	   sep=argv[i][j];
	   argv[i][j]=0;               /* create a 'symbol' for the parsed */
				       /* 'macro' name                     */
	   mname=mk_symbol(strdup(argv[i]+2),IDENT,0l);
	   if(argv[i][1]=='D')         /* process '-D' flag (followed by   */
				       /* identifier 'mname' and the       */
				       /* character 'sep')                 */
	   {
	     symbol *replc;                 /* 'symbol' list specifying    */
					    /* the 'macro' replacement     */
	     replc=(symbol *)               /* build NULL terminated list  */
		   malloc(2*sizeof(symbol));/* with one element            */
	     replc[0]=replc[1]=NULL;
	     switch(sep)
	     {
	       case '=': replc[0]=mk_symbol(strdup(argv[i]+j+1),PPCOTHER,0l);
	       case 0  :		    /* define macro                */
			 if(definem(mname,NULL,replc,0)) break;
	       default : usage('D');        /* generate error on illegal   */
	     }                              /* 'sep' character or illegal  */
	   }                                /* 'macro' redefinition        */
	   else                        /* process '-U' flag (followed by a */
				       /* identifier 'mname' and the       */
	     if(!sep)                  /* character 'sep')                 */
	       undefm(mname);               /* undefine 'macro'            */
	     else
	       usage('U');                  /* generate error on illegal   */
					    /* 'sep' character (must be 0  */
	   break;                           /* to terminate the param.text)*/
      default  :
	   usage(argv[i][1]);          /* exit with error: flag was wether */
    }                                  /* '-D' nor '-U'                    */

  }                                    /* continue with next flag (loop)   */

  return(dest);                        /* return with 'dest' 'stream'      */
}                                      /* created before (step 7)          */
