/***************************************************************************/
/*                                                                         */
/* IDENTITY:         evalexpr                                              */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  05.04.91                */
/*                   CODE             Schlotfeldt  30.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:02         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (in LINE_EXP mode, see module 'expandm')              */
/*                   'evalexpr' assumes the stream to describe a constant  */
/*                   expresssion using long/unsigned long arithmetic.      */
/*                   Preceding/succeeding WHITESPace will be read too.     */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'const_expr'                        */
/*                   for local use   : 'star', 'slash', 'pcent', 'plus',   */
/*                                     'minus', 'shiftr', 'shiftl', 'less',*/
/*                                     'greater', 'lsorequ', 'groeequ',    */
/*                                     'notequ', 'equal', 'binand', 'bin-  */
/*                                     or', 'binxor', 'logand', 'logor',   */
/*                                     'eval_const', 'eval0' ... 'eval11'  */
/*                                                                         */
/***************************************************************************/
#include "evalexpr.h"

#include "stringu.h"
#include "lexbuf.h"
#include "expandm.h"

#include "errmsg.h"
#include "myctype.h"

typedef struct                         /* local type for numbers repre-    */
	{                              /* sented by their value 'v' and    */
	  unsigned long v;             /* their type.                      */
	  char          u;             /* numbers with 'u' part==0 will be */
	}                              /* interpreted as long (done by     */
	value;                         /* type casting) others are taken   */
				       /* as unsigned long                 */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         STERM                                                 */
/*                                                                         */
/* TYPE:             Local Macro                                           */
/*                                                                         */
/* FUNCTION:         Defines a c-function of type 'value' parsing a list   */
/*                   of sub-terms of a constant expression separated one   */
/*                   of 3 different 'symbols' (operators).                 */
/*                   The sub-terms are evaluated (parsed) by another spec. */
/*                   c-function returning a 'value' for this sub-term.     */
/*                                                                         */
/* PARAMETERS:       1) name of function to define                         */
/*                   2) name of function to use for sub-term parsing       */
/*                   3) name of 1st separator 'symbol'                     */
/*                   4) name of function to execute if actual separator was*/
/*                      the 'symbol' spec. by parameter 3                  */
/*                   5+6) same as 3+4) for the 2nd separator 'symbol'      */
/*                   7+8) same as 3+4) for the 3rd separator 'symbol'      */
/*                                                                         */
/* RETURN VALUE:     return value of the defined function is the 'value'   */
/*                   represented by the parsed term and a flag (set on     */
/*                   error)                                                */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'symbols' to parse are read by 'next_lex_e' in LINE-  */
/*                   EXP mode. The internal loop for list processing will  */
/*                   exit on error or first illegal separator/operator     */
/*                   'symbol' which will be written back to the 'lex_buf'  */
/*                    by 'uget1_lex'                                       */
/*                   When the error flag (passed to all functions) is set, */
/*                   the defined function will exit immediately.           */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag won't be set directly by a function    */
/*                   defined by this macro. But calls to sub-functions     */
/*                   can set the flag signaling an error.                  */
/*                                                                         */
/***************************************************************************/

#define STERM(af,pf,s1,f1,s2,f2,s3,f3) \
 value af S1(char *,errflg)            \
 {                                     \
   value  a;                           /* temporary storage for a 'value'  */\
   symbol p;                           /* temporary storage for a 'symbol' */\
				       \
   a=pf(errflg);                       /* evaluate 1st sub-term and exit   */\
   if(*errflg) return(a);              /* on error                         */\
				       \
   while(!*errflg)                     /* process further pairs of separa- */\
   {                                   /* tor 'symbols' and sub-terms      */\
     p=next_lex_e(LINE_EXP);           /* get next 'symbol' and compare it */\
				       /* with the known separators (opera-*/\
				       /* tors)                            */\
	  if(p==s1)                    \
	    a=f1(a,pf(errflg),errflg); \
     else if(p==s2)                    \
	    a=f2(a,pf(errflg),errflg); \
     else if(p==s3)                    \
	    a=f3(a,pf(errflg),errflg); \
     else break;                       /* exit loop on unknown operator    */\
   }                                   \
   uget1_lex(p);                       /* write back one 'symbol' to the   */\
   return(a);                          /* 'lex_buf' and return calculated  */\
 }                                     /* 'value'                          */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         DEF_OP                                                */
/*                                                                         */
/* TYPE:             Local Macro                                           */
/*                                                                         */
/* FUNCTION:         Defines a c-function which calculates the result of   */
/*                   applying a specified operator to 2 'values' (in con-  */
/*                   sideration of their types!)                           */
/*                                                                         */
/* PARAMETERS:       1) name of function to define                         */
/*                   2) operator (normal c style)                          */
/*                   3) Flag: 1=type casting is done before operation      */
/*                              (result is long if both arguments are long;*/
/*                              otherwise the result is unsigned long)     */
/*                            0=no type casting                            */
/*                              (result has the same type as the 1st argu- */
/*                              ment)                                      */
/*                   4) Flag: 1=second argument is checked not be zero,    */
/*                              else -> error                              */
/*                            0=no zero checking                           */
/*                                                                         */
/* RETURN VALUE:     return value of the defined function is the           */
/*                   calculated 'value' and the (modified) error flag.     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag may be set if 4th parameter was 1 and  */
/*                   the second argument to the defined function was 0.    */
/*                   This will generate the error message "divide by zero".*/
/*                                                                         */
/***************************************************************************/
#define DEF_OP(name,op,cst,dbz)        \
 value name S3(value,a,value,b,        \
	       char *,errflg)          \
 {                                     \
   if(cst) a.u=b.u=(a.u||b.u);         /* perform type casting if desired  */\
   if(dbz && !b.v)                     /* check 2nd argument for zero      */\
   {                                   \
     errmsg(1,16,NULL,NULL);           /* generate error message and set   */\
     *errflg=1;                        /* the error flag                   */\
   }                                   \
   else                                \
     a.v                               /* calculate new 'v' part for the   */\
     =a.v op ((b.u)?(unsigned long)b.v /* 'value' to return                */\
		  :(long)         b.v);\
   return(a);                          \
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         star   ,slash  ,pcent  ,plus   ,minus  ,shiftr ,      */
/*                   shiftl ,greater,grorequ,less   ,leorequ,notequ ,      */
/*                   equal  ,binand ,binor  ,binxor ,logand ,logor         */
/*                                                                         */
/* TYPE:             Local Functions                                       */
/*                                                                         */
/* FUNCTION:         Calculate the result of applying a specific operator  */
/*                   to 2 'values' (in consideration of their types!)      */
/*                                                                         */
/* PARAMETERS:       1) 1st argument of term to calculate                  */
/*                   2) 2nd argument of term to calculate                  */
/*                   3) error flag                                         */
/*                                                                         */
/* RETURN VALUE:     calculated 'value' and the (modified) error flag.     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag may be set for functions 'slash' and   */
/*                   'pcent' if the 2nd argument was 0.                    */
/*                   This will generate the error message "divide by zero".*/
/*                                                                         */
/***************************************************************************/
DEF_OP(star   , *,1,0)
DEF_OP(slash  , /,1,1)
DEF_OP(pcent  , %,1,1)
DEF_OP(plus   , +,1,0)
DEF_OP(minus  , -,1,0)
DEF_OP(shiftr ,>>,0,0)
DEF_OP(shiftl ,<<,0,0)
DEF_OP(greater, >,1,0)
DEF_OP(grorequ,>=,1,0)
DEF_OP(less   ,< ,1,0)
DEF_OP(leorequ,<=,1,0)
DEF_OP(notequ ,!=,1,0)
DEF_OP(equal  ,==,1,0)
DEF_OP(binand , &,1,0)
DEF_OP(binor  , |,1,0)
DEF_OP(binxor , ^,1,0)
DEF_OP(logand ,&&,1,0)
DEF_OP(logor  ,||,1,0)

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval_const                                            */
/*                                                                         */
/* TYPE:             Local Function                                        */
/*                                                                         */
/* FUNCTION:         Evaluates a elementary constant read as 'symbol' via  */
/*                   'next_lex_e' (legal are character constants, integer  */
/*                   constants and identifiers)                            */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of elementary constant and the (modified)     */
/*                   error flag.                                           */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Switches on the type of the first 'symbol' read via   */
/*                   'next_lex_e'. Legal types are CCONST, ICONST or IDENT.*/
/*                   Character CONSTants are evaluated to their signed     */
/*                   ASCII value (provided that the constant doesn't re-   */
/*                   present a w_char type character, recognized by the    */
/*                   leading L)                                            */
/*                   Integer CONSTants are converted by 'str2ul' (module   */
/*                   'stringu'). They are long by default. Interpretation  */
/*                   as unsigned long is forced by using the u modifier    */
/*                   or done automatic if long representation would pro-   */
/*                   duce a negativ value.                                 */
/*                   IDENTifiers are evaluated to 0.                       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag is set on any error (eg. illegal type  */
/*                   of constant)                                          */
/*                                                                         */
/***************************************************************************/
value eval_const S1(char *,errflg)
{
  value  r;                            /* result of 'eval_const'           */
  symbol p;                            /* temporary storage for a 'symbol' */
  char   *s;                           /* pointer used for converting      */
				       /* 'symbol' text to a numerical     */
				       /* value                            */
  char   ef=0;                         /* local error flag                 */

  p=next_lex_e(LINE_EXP);              /* get 'symbol' representing the    */
  switch(p->type) 		       /* constant                         */
  {
    case CCONST :                      /* analyze Character CONSTant       */
	 r.u=0;                             /* result is signed            */
	 s=p->text;                         /* constant mustn't use w_char */
	 if(*s++!='\'')  ef=1;              /* characters                  */
	 r.v=strtoc(&s,0,&ef);              /* evaluate result             */
	 if(*s++!='\'')  ef=1;              /* check errors on string con- */
	 if(*s)          ef=1;              /* vertion                     */
	 break;
    case ICONST :                      /* analyze Integer CONSTant         */
	 r.v=str2ul(p->text,&s,0);          /* evaluate result             */
	 if(s==p->text)  ef=1;              /* check errors on string con- */
					    /* vertion                     */
	 if(tolower(*s)=='l') s++;          /* skip l modifier on 'symbol' */
					    /* text                        */
	 r.u=((long)r.v<0                   /* set right type of value     */
	      || tolower(*s)=='u');         /* (long or unsigned long)     */
	 break;
    case IDENT  :                      /* analyze IDENTifier               */
	 r.u=r.v=0;                         /* evaluates always to 0       */
	 break;
    default     :                      /* set local error flag on any other*/
	 ef=1;                         /* 'symbol' type                    */
  }
  if(ef)                               /* generate error message ("illegal */
  {                                    /* constant") and set error flag if */
    errmsg(1,35,NULL,NULL);            /* local error flag has been set    */
    *errflg=1;
    uget1_lex(p);
  }
  return(r);                           /* return with result               */
}                                      

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval11                                                */
/*                                                                         */
/* TYPE:             Local Function                                        */
/*                                                                         */
/* FUNCTION:         Evaluates a constant read as sequence of 'symbols' via*/
/*                   'next_lex_e'. Legal are elementary constants (see     */
/*                   'eval_const') or whole constant expressions enclosed  */
/*                   by parenthesis.                                       */
/*                   Preceding and succeeding WHITESPace will be skipped.  */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of constant and the (modified) error flag.    */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag can be set by any subsequent function  */
/*                   or on a missing right parenthesis. In this case       */
/*                   'eval11' generates an error message.                  */
/*                                                                         */
/***************************************************************************/
value eval11 S1(char *,errflg)
{
  value  a,                            /* result of 'eval11'               */
	 eval0 E1(char *);             /* forward declaration of 'eval0'   */
  symbol p;                            /* temporary storage for a 'symbol' */

  p=next_lex_e(LINE_EXP);              /* get next 'symbol' and skip       */
  skipsp(p,LINE_EXP)                   /* WHITESPace                       */
  if(p==L_PARENTH)                     /* evaluate whole const. expression */
  {                                    /* by calling 'eval0' if actual     */
    a=eval0(errflg);                   /* 'symbol' is '('                  */
    p=next_lex_e(LINE_EXP);            /* the next 'symbol' must be ')'    */
    if(p!=R_PARENTH)                   /* else: -> error                   */
    {
      errmsg(1, 6,NULL,NULL);
      *errflg=1;
      uget1_lex(p);
    }
    else
      p=next_lex_e(LINE_EXP);          /* skip ')'                         */
  }
  else
  {
    uget1_lex(p);                      /* call 'eval_const' if 'symbol'    */
    a=eval_const(errflg);              /* read was not '('                 */
    p=next_lex_e(LINE_EXP);
  }
  skipsp(p,LINE_EXP)                   /* skip WHITESPace and write back   */
  uget1_lex(p);                        /* first non-WHITESPace 'symbol'    */
  return(a);                           /* before returning the result      */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval10                                                */
/*                                                                         */
/* TYPE:             Local Function                                        */
/*                                                                         */
/* FUNCTION:         Evaluates a signed constant read as sequence of       */
/*                   'symbols' via 'next_lex_e'. Legal are constants (see  */
/*                   'eval11') or constants preceded by 1 of the following */
/*                   'symbols': PLUS, MINUS, NOT, COMPLM                   */
/*                   Preceding WHITESPace will be skipped.                 */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of signed constant and the (modified) error   */
/*                   flag.                                                 */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag won't be set directly by 'eval10' .    */
/*                   But calls to sub-functions can set the flag signaling */
/*                   an error.                                             */
/*                                                                         */
/***************************************************************************/
value eval10 S1(char *,errflg)
{
  value  a;                            /* result of 'eval10'               */
  symbol p;                            /* temporary storage for a 'symbol' */

  p=next_lex_e(LINE_EXP);              /* get next non-WHITESPace 'symbol' */
  skipsp(p,LINE_EXP)

       if(p==PLUS     ) { a=eval11(errflg);            }
  else if(p==MINUS    ) { a=eval11(errflg); a.v= -a.v; }
  else if(p==NOT      ) { a=eval11(errflg); a.v= !a.v; }
  else if(p==COMPLM   ) { a=eval11(errflg); a.v= ~a.v; }
  else   { uget1_lex(p);  a=eval11(errflg);            }
  return(a);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval9 to eval2                                        */
/*                                                                         */
/* TYPE:             Local Functions                                       */
/*                                                                         */
/* FUNCTION:         Evaluation of sub-terms read as sequence of 'symbols' */
/*                   via 'next_lex_e'. Legal are sub-terms using operators */
/*                   of a specific precedence level.                       */
/*                   'eval9'/'eval1') interpretes the operators with the   */
/*                   highest/lowest precedence.                            */
/*                   So 'eval1' calls 'eval2', 'eval2' calls 'eval3' etc.  */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of sub-term and the (modified) error flag     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Functions are defined by using the 'STERM' macro.     */
/*                   Look at the definition of this macro for detailed     */
/*                   information.                                          */
/*                   Functions processing just 2 different operators du-   */
/*                   plicate their last pair of arguments to the 'STERM'   */
/*                   macro. This will produce a little bit useless code,   */
/*                   but it works correct.                                 */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: See 'STERM' macro                                     */
/*                                                                         */
/***************************************************************************/
STERM(eval9 ,eval10, MULTIPL  ,star   ,SLASH    ,slash  ,MODULO   ,pcent  )
STERM(eval8 ,eval9 , PLUS     ,plus   ,MINUS    ,minus  ,MINUS    ,minus  )
STERM(eval7 ,eval8 , R_SHIFT  ,shiftr ,L_SHIFT  ,shiftl ,L_SHIFT  ,shiftl )
STERM(eval6 ,eval7 , GREATER  ,greater,GROREQUAL,grorequ,GROREQUAL,grorequ)
STERM(eval5 ,eval6 , LESS     ,less   ,LSOREQUAL,leorequ,LSOREQUAL,leorequ)
STERM(eval4 ,eval5 , NOT_EQUAL,notequ ,EQUAL    ,equal  ,EQUAL    ,equal  )
STERM(eval3 ,eval4 , BIN_AND  ,binand ,BIN_OR   ,binor  ,BIN_XOR  ,binxor )
STERM(eval2 ,eval3 , LOG_AND  ,logand ,LOG_OR   ,logor  ,LOG_OR   ,logor  )

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval1                                                 */
/*                                                                         */
/* TYPE:             Local Function                                        */
/*                                                                         */
/* FUNCTION:         Same function as 'eval9' to 'eval2'.                  */
/*                   This function processes the ?: operator.              */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of sub-term and the (modified) error flag     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'eval2' to evaluate the first sub-term. If      */
/*                   the next 'symbol' is the QUESTionMark, there must     */
/*                   follow 2 further sub-terms (evaluated recursively by  */
/*                   this function) separated by a COLON.                  */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag can be set by any subsequent function  */
/*                   or on a missing colon. In this case 'eval1' generates */
/*                   an error message.                                     */
/*                                                                         */
/***************************************************************************/
value eval1 S1(char *,errflg)
{
  value  a,b,c;                        /* storage for the results of the   */
				       /* 3 sub-terms                      */
  symbol p;                            /* temporary storage for a 'symbol' */

  a=eval2(errflg);                     /* evaluate 1st sub-term and        */
  p=next_lex_e(LINE_EXP);              /* check next 'symbol' to be a      */
  if(p==QUESTM)                        /* QUESTionMark                     */
  {
    b=eval1(errflg);                   /* evaluate 2. sub-term recursively */
    p=next_lex_e(LINE_EXP);            /* next 'symbol' must be a colon!   */
    if(p==COLON)
    {
      c=eval1(errflg);                 /* evaluate 3. sub-term recursively */
      return(a.v?b:c);                 /* and return result                */
    }
    else
    {
      errmsg(1,17,NULL,NULL);          /* generate an error message for    */
      *errflg=1;                       /* the missing colon                */
    }
  }
  uget1_lex(p);                        /* write back last 'symbol' and     */
  return(a);                           /* take 'a' as result in case of    */
}                                      /* error or no ?: operator          */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         eval0                                                 */
/*                                                                         */
/* TYPE:             Local Function                                        */
/*                                                                         */
/* FUNCTION:         Same function as 'eval9' to 'eval1'.                  */
/*                   This function processes the ',' operator. The inter-  */
/*                   pretable input to this function is already a whole    */
/*                   constant expression!                                  */
/*                                                                         */
/* PARAMETERS:       error flag                                            */
/*                                                                         */
/* RETURN VALUE:     'value' of constant expression and the (modified)     */
/*                   error flag                                            */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'eval1' to evaluate the first sub-term. Further */
/*                   sub-terms are evaluated (overwriting the first result)*/
/*                   if the last sub-term is followed by a COMMA.          */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: The error flag won't be set directly by 'eval0'.      */
/*                   But calls to sub-functions can set the flag signaling */
/*                   an error.                                             */
/*                                                                         */
/***************************************************************************/
value eval0 S1(char *,errflg)
{
  value  a;                            /* result of 'eval10'               */
  symbol p;                            /* temporary storage for a 'symbol' */

  do                                   /* repeat evaluation of sub-terms   */
  {                                    /* by calling 'eval1' til next      */
    a=eval1(errflg);                   /* 'symbol' isn't a COMMA or error  */
    p=next_lex_e(LINE_EXP);            /* flag has been set                */
  }
  while(p==COMMA && !*errflg);
  uget1_lex(p);                        /* write back last 'symbol' and     */
  return(a);                           /* return the 'value' of the last   */
}                                      /* sub-term evaluated by 'eval1'    */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         const_expression                                      */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Same function as 'eval0' (includes storage for the    */
/*                   error flag used by all functions above)               */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     'value' of constant expression or (explicit) 0 on     */
/*                   error.                                                */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'eval0'                                         */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Returns 0 on error. (This number isn't a legal line   */
/*                   number (ppLINE) and will be interpreted as FALSE by   */
/*                   'if_sect')                                            */
/*                                                                         */
/***************************************************************************/
unsigned long const_expr S0(void)
{
  value  a;
  char   errflg=0;

  a=eval0(&errflg);
  return(errflg?0:a.v);
}
