/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ifsect                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  05.04.91                */
/*                   CODE             Schlotfeldt  01.05.91                */
/*                   VERSION(CURRENT) Schlotfeldt  01.06.91  16:55         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm').                                   */
/*                   'ifsect' assumes the stream to describe a pp-command  */
/*                   line, which was already identified as #if/#if(n)def-  */
/*                   line by reading the first 'symbol' of this line by    */
/*                   the 'parser' module.                                  */
/*                   To parse lines between the #if/#if(n)def-line end the */
/*                   attached #else/#elif/#endif-lines this module will    */
/*                   call recursively the 'parser'.                        */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'if_section'                        */
/*                                                                         */
/***************************************************************************/
#include "ifsect.h"

#include "parser.h"
#include "evalexpr.h"
#include "macro.h"
#include "expandm.h"

#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         if_section                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         complete parser to a 'if_section' including all sub-  */
/*                   groups (separated by #else/#elif-lines, terminated by */
/*                   a #endif-line)                                        */
/*                   Further informations: see module header               */
/*                                                                         */
/* PARAMETERS:       1) last 'symbol' read by the 'parser'                 */
/*                      (must be one of the following: IF, IFDEF or IFNDEF)*/
/*                   2) actual 'stream' to use for output or NULL if out-  */
/*                      put and pp-command line processing (excluding      */
/*                      #else/#elif- and #endif-lines) is disabled (see    */
/*                      parameters to 'group' in module 'parser')          */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal syntax of   */
/*                   the 'if_section' (missing #endif-line, multiple #else-*/
/*                   lines or illegal macro-name (not a IDENTifier) on     */
/*                   #ifdef/#ifndef-lines).                                */
/*                                                                         */
/***************************************************************************/
void if_section S2(symbol,p,stream *,outf)
{
# define reterr(n)                     /* reterr:                          */\
  {                                    /*   generate error message no. 'n',*/\
    errmsg(1,n,NULL,NULL);             /*   ignore remaining 'symbols' on  */\
    skip2eol(p,NO_EXP)                 /*   the actual line and return to  */\
    return;                            /*   calling function ('group')     */\
  }

  char bool,                           /* boolean flag for the result of   */
				       /* the last if/elif-clause          */
       nogr=1;                         /* flag: cleared if 'bool' evaluates*/
				       /* to 1, the cleared flag disables  */
				       /* futher setting of 'bool'         */
				       /* (only 1 group in a 'if_section'  */
				       /* may be active)                   */

  do                                  /* loop repeated til ELSE, ENDIF or */
  {                                    /* ENDOFFILE is reached (processes  */
				       /* IF/ELIF groups)                  */
    bool=(p==IF || p==ELIF)
	 ?(const_expr()!=0)            /* evaluate constant expression to  */
				       /* set 'bool' (for act.'symbol' is  */
				       /* IF or ELIF)                      */
	 :(p==IFNDEF);                 /* or store information wether act. */
				       /* 'symbol' is IFNDEF (1) or IFDEF  */
				       /* (0)                              */
    if(p==IFDEF || p==IFNDEF)
    {
      p=next_lex_e(NO_EXP);            /* get next non-WHITESPace 'symbol' */
      skipsp(p,NO_EXP)                 /* from the #ifdef/#ifndef-line     */
      if(p->type!=IDENT) reterr( 7)    /* must be an IDENTifier            */
      bool^=(searchm(p)!=NULL);        /* set 'bool' if IDENTifier was     */
				       /* defined as 'macro'               */
				       /* 'bool'-result is negated on      */
				       /* #ifndef-line (previous value of  */
				       /* 'bool' was 1)                    */
    }
    if(bool&=nogr) nogr=0;
    p=next_lex_e(NO_EXP);              /* next 'symbol' must terminate the */
    if(!endofline(p)) reterr(23)       /* line                             */
    group(bool?outf:NULL,1);           /* call 'group' recursively to parse*/
				       /* the next group of lines          */
    p=next_lex_e(NO_EXP);              /* continue with the next 'symbol'  */
  }                                    /* (terminating the parsed group:   */
  while(p==ELIF);                      /* ELIF, ELSE, ENDIF or ENDOFFILE)  */

  if(p==ELSE)                          /* process #else group:             */
  {
    p=next_lex_e(NO_EXP);                   /* next 'symbol' must terminate*/
    if(!endofline(p)) reterr(23)            /* the line                    */
    group(nogr?outf:NULL,1);                /* call 'group' recursively to */
					    /* parse the next group of     */
					    /* lines                       */
    p=next_lex_e(NO_EXP);                   /* get next 'symbol' (should   */
  }                                         /* be ENDIF!)                  */

  if(p==ENDIF)                         /* process 'if_section' termination:*/
  {
    p=next_lex_e(NO_EXP);                   /* next 'symbol' must terminate*/
    if(!endofline(p)) reterr(23)            /* the line                    */
    return;
  }
  reterr(24)                           /* generate error for missing ENDIF */
}

