/***************************************************************************/
/*                                                                         */
/* IDENTITY:         lexbuf                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  16.03.91                */
/*                   CODE             Schlotfeldt  12.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  09.06.91  19:50         */
/*                                                                         */
/* FUNCTION:         Realizes buffered input of 'symbols' via 'next_lex'.  */
/*                   'symbols' can be fetched, written back (explicit),    */
/*                   written back all at once (implicit) and reduced       */
/*                   (= clearing the internal buffer)                      */
/*                                                                         */
/* DATA/TABLES:      Contains static data (buffer and pointer to the       */
/*                   buffer) hidden in the not global used function        */
/*                   'lex_buf_op'                                          */
/*                                                                         */
/* PROCEDURES:       for external use: 'get_lex', 'uget1_lex',             */
/*                                     'uget_all_lex', 'clear_lex',        */
/*                                     'init_lexbuf', 'uget_lexl'          */
/*		     for local use   : 'lex_buf_op'                        */
/*                                                                         */
/***************************************************************************/
#include "lexbuf.h"

#include "lexi.h"

#include "myalloc.h"
#include "errmsg.h"

enum cmd_t                             /* local type for calls to          */
     { get_,uget_all_,                 /* 'lex_buf_op'                     */
       uget1_,clear_,init_ };

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         lex_buf_op                                            */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         'init_' function:                                     */
/*                     Initialize the local static data (buffer and ptr.)  */
/*                   'get_' function:                                      */
/*                     Gets the next 'symbol' from buffer or 'next_lex' if */
/*                     end of buffer is reached. In this case the return   */
/*                     value will be 1 else 0.                             */
/*                   'uget1_' function:                                    */
/*                     Un-gets one 'symbol' specified by 2nd parameter     */
/*                     previous contents of buffer (at the position to     */
/*                     read 'symbols' from via 'get_') will be lost.       */
/*                   'uget_all_' function:                                 */
/*                     Un-gets all 'symbols' read via 'get_' since last    */
/*                     call of function 'init_' or 'clear_'. The contents  */
/*                     of the buffer in the attached positions is left     */
/*                     unchanged. 'uget_all_' function includes 'clear_':  */
/*                   'clear_' function:                                    */
/*                     Clears internal buffer and sets new mark for 'uget- */
/*                     all_'. Calls to function 'uget1_' will be possible  */
/*                     on a cleared buffer too.                            */
/*                                                                         */
/* PARAMETERS:       1) Command to specify one of the 5 actions            */
/*                   2) Symbol for 'uget1_' function                       */
/*                                                                         */
/* RETURN VALUE:     modified 'symbol' parameter (return value for         */
/*                   'get_' function)                                      */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  First call to this routine should activate the 'init_'*/
/*                   function. Don't use the un-initialized routine!       */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' if there are more 'symbols' to buffer  */
/*                   than the maximum value for the machine dependend type */
/*                   'size_t'/'sizeof(symbol)'!                            */
/*                                                                         */
/***************************************************************************/
void lex_buf_op S2(enum cmd_t,cmd,symbol *,s)
{
# define BLKSIZE 128                   /* minimum size for memory blocks   */
				       /* allocated by 'lex_buf_op'        */
				       /* high values minimize calls to    */
				       /* 'realloc' and moving of buffer   */
				       /* contents                         */
				       /* should be a power of 2 !!!       */

  static symbol *buffer;               /* buffer for 'symbols'             */
  static size_t fpos,                  /* pos. of first valid entry        */
		lpos,                  /* last allocated pos.+1            */
		rpos,                  /* pos. to read from                */
		wpos;                  /* pos. to write to                 */

  auto   size_t i;                     /* loop variable                    */

  switch(cmd)
  {
    case init_:                        /* initialize local static data     */
	 buffer=NULL;
	 fpos=lpos=rpos=wpos=0;
	 return;

    case get_:                         /* get next 'symbol'                */
	 if(rpos>=wpos)
	 {
	   if(wpos>=lpos)
	   {
	     size_t new;

	     new=(lpos+=BLKSIZE)*sizeof(symbol);
	     if(new<=0) errmsg(1,38,NULL,NULL);
	     buffer=(symbol *)realloc(buffer,new);
	   }
	   buffer[rpos=wpos++]
	     =next_lex(0);             /* "fresh" or...                    */
	 }
	 *s=buffer[rpos++];            /* ...buffered                      */
	 return;

    case uget1_:                       /* unget 1 spec. 'symbol':          */
	 if(rpos>fpos)
	 {                             /* put 'symbol' on decremented      */
	   buffer[--rpos]= *s;         /* 'rpos' if buffer isn't empty     */
	   return;
	 }                             /* else:                            */
	 if(fpos)                      /* use allocated space of lower     */
	 {                             /* buffer position. update 'rpos'   */
	   buffer[rpos= --fpos]= *s;   /* and 'fpos'                       */
	   return;
	 }                             /* else (there is no space):        */
	 buffer=                       /* 'realloc' buffer and shift       */
	   (symbol *)                  /* contents one position right...   */
	   realloc(buffer,++lpos*sizeof(symbol));
	 if(lpos<=0) errmsg(1,38,NULL,NULL);
	 for(i=wpos++;i>0;i--)
	   buffer[i]=buffer[i-1];
	 buffer[rpos=fpos=0]= *s;      /* ...to insert 'symbol' on the new */
	 return;                       /* position                         */

    case uget_all_:                    /* unget all 'symbols' fetched since*/
	 rpos=fpos;                    /* last call of 'init_' or 'clear_' */
	 return;                       /* function (resets 'rpos')         */

    case clear_:                       /* clear buffer:                    */
	 if((fpos=rpos)>=BLKSIZE)      /* logical clearing by updating     */
				       /* 'fpos' or more physical clearing */
	 {                             /* by moving the contents of the    */
	   lpos-=fpos;                 /* buffer (reducing unused space    */
	   for(i=1;i<lpos;i++)         /* in the lower (left) positions)   */
	     buffer[i-1]=buffer[fpos+i-1];
	   wpos-=fpos;
	   fpos=rpos=0;
	 }
	 return;
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_lex                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets next 'symbol' from the implemented 'symbol'      */
/*                   buffer (see 'lex_buf_op)                              */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     'symbol' fetched from buffer/'next_lex'               */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'lex_buf_op', return value is stored in the     */
/*                   modified parameter to 'lex_buf_op'                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Before you use 'get_lex' this module must be          */
/*                   initialized by 'init_lexbuf'                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
symbol get_lex S0(void)
{
  symbol p;

  lex_buf_op(get_,&p);
  return(p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         uget_all_lex                                          */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Ungets all 'symbols' read via 'get_lex' since last    */
/*                   call to 'clear_lex', 'init_lexbuf' or 'uget_all_lex'  */
/*                   ('uget_lex_all' includes 'clear_lex' function)        */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'lex_buf_op'                                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Before you use 'uget_all_lex' this module must be     */
/*                   initialized by 'init_lexbuf'                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void uget_all_lex S0(void)
{
  lex_buf_op(uget_all_,NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         uget1_lex                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Ungets one specific 'symbol' to the implemented buffer*/
/*                   (see 'lex_buf_op'). This 'symbol' will perhaps over-  */
/*                   write a buffered 'symbol'                             */
/*                                                                         */
/* PARAMETERS:       'symbol' to write back                                */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'lex_buf_op' with passed parameter              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Before you use 'uget1_lex' this module must be        */
/*                   initialized by 'init_lexbuf'                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void uget1_lex S1(symbol,p)
{
  lex_buf_op(uget1_,&p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         uget_lexl                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Ungets a list of 'symbols' to the implemented buffer  */
/*                   (see 'lex_buf_op'). This 'symbols' will perhaps over- */
/*                   write buffered 'symbols'                              */
/*                   The list will be processed back to front! Calls with  */
/*                   empty list or NULL are ignored.                       */
/*                                                                         */
/* PARAMETERS:       NULL terminated 'symbol' list to write back           */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls repeatedly 'uget_lexl'for each 'symbol' in the  */
/*                   list. The list will be set 'free' after use.          */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Before you use 'uget_lexl' this module must be        */
/*                   initialized by 'init_lexbuf'                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void uget_lexl S1(symbol *,p)
{
  size_t i=0;                          /* length of 'symbol' list          */

  if(p)
  {
    while(p[i]) i++;                   /* count 'symbols' in the list      */
    while(i)
      lex_buf_op(uget1_,&p[--i]);      /* write back 'symbols'             */
    free(p);
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         clear_lex                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Clears the implemented buffer (see 'lex_buf_op') and  */
/*                   sets a new mark for 'uget_all_lex'                    */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'lex_buf_op'                                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Before you use 'clear_lex' this module must be        */
/*                   initialized by 'init_lexbuf'                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void clear_lex S0(void)
{
  lex_buf_op(clear_,NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_lexbuf                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initialization of this module (see 'lex_buf_op') and  */
/*                   'lexi' module. Parameters are just used for 'init-    */
/*                   lexi' call.                                           */
/*                                                                         */
/* PARAMETERS:       Maximum (significant) length of an identifier or 0    */
/*                   for no limitation.                                    */
/*                   Flag for recognition of nested comments (0=off).      */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'lex_buf_op' and 'init_lexi' with passed        */
/*                   parameters (unchanged)                                */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  This function should be called as first one of        */
/*                   this module!                                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void init_lexbuf S3(long,midl,char,nest,char,removeleadingblanks)
{
  lex_buf_op(init_,NULL);
  init_lexi(midl,nest,removeleadingblanks);
}
