/***************************************************************************/
/*                                                                         */
/* IDENTITY:         mergestr                                              */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  01.04.91                */
/*                   CODE             Schlotfeldt  25.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:22         */
/*                                                                         */
/* FUNCTION:         Filter used by the parser instead of direct calls     */
/*		     to 'output' module.                                   */
/*                   Manages concatenation of strings (separated by any    */
/*                   kind of white space including '\n') before output is  */
/*                   done by the 'output' module.                          */
/*                   The first call to this module should be a initiali-   */
/*                   zation call to 'merge_strings' (use 'init_mergestr'   */
/*                   macro as defined in 'mergestr.h' !).                  */
/*                                                                         */
/* DATA/TABLES:      Contains static data (buffer for lines ending on a    */
/*                   string -> concatenation may be possible)              */
/*                                                                         */
/* PROCEDURES:       for external use: 'merge_strings'                     */
/*                   for local use   : 'merge_str', 'merge_1line',         */
/*                                     'mrgstr_outp'                       */
/*                                                                         */
/***************************************************************************/
#include "mergestr.h"

#include "symbol.h"
#include "strinfo.h"
#include "output.h"

#include "stringu.h"
#include "myalloc.h"
#include "errmsg.h"

typedef struct                         /* structure for all informations   */
	{                              /* needed to the whole output for a */
				       /* line:                            */
	  stream *outf;                /*   'stream' to use for output     */
	  symbol *cont;                /*   contents of line (as 'symbol'  */
				       /*   list)                          */
	  size_t len;                  /*   number of 'symbols' in the     */
				       /*   above list                     */
	  char   *fname,               /*   name and line number of the    */
		 *linen,               /*   actual source (see 'strinfo')  */
		 *nlnno;	       /*   same as 'linen' for the next   */
				       /*   line                           */
	 } line_t;

#define free_line(l)                   /* free_line:                       */\
 {                                     /*   sets all dynamic allocated     */\
   free((l)->linen);                   /*   space for a variable pointing  */\
   free((l)->fname);                   /*   to a line_t structure 'free'   */\
   free((l)->nlnno);                                                           \
   free((l)->cont );                                                           \
 }

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         merge_str                                             */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Tries to merge 2 strings (given as 'symbols') to a    */
/*                   new string.                                           */
/*                                                                         */
/* PARAMETERS:       2 'symbols' (of type STRING) which can be possibly    */
/*                   concatenated                                          */
/*                                                                         */
/* RETURN VALUE:     New 'symbol' representing the concatenated string or  */
/*                   NULL if concatenation was impossible.                 */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   String merging is only possible, if both strings      */
/*                   start with same character (usually quotes).           */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Parameter 'symbols' must be previously checked to be  */
/*                   STRINGs                                               */
/*                                                                         */
/* ERROR CONDITIONS: Returns NULL on impossible merging (additional call   */
/*                   to 'errmsg' if the result string would be longer than */
/*                   the maximum value for the machine dependend type      */
/*                   'size_t' !                                            */
/*                                                                         */
/***************************************************************************/
symbol merge_str S2(symbol,s1,symbol,s2)
{
  size_t l1,l2;                        /* lengths of the original strings  */
  char   *s;                           /* new string                       */

  if(s1->text[0]!=s2->text[0])         /* check if strings use the same    */
    return(NULL);                      /* character type (char or w_char)  */
  l1=strlen(s1->text);
  l2=strlen(s2->text);
  if(l1+l2<=0)
  {
    errmsg(1,40,NULL,NULL);
    return(NULL);
  }
  s=(char *)                           /* allocate space for the new string*/
    malloc(l1+l2-((s2->text[0]=='L')?2:1));
  strcpy(s,s1->text);                  /* build new string                 */
  s[l1-1]=0;
  strcat(s,s2->text+((s2->text[0]=='L')?2:1));
  return(mk_symbol(s,STRING,0l));      /* create new 'symbol' (STRING) and */
}                                      /* return it to the calling function*/

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         merge_1line                                           */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Tries to merge STRING 'symbols' on the current line   */
/*                                                                         */
/* PARAMETERS:       List of 'symbols' representing one line of the        */
/*                   (modified) source text as generated by the parser     */
/*                   for output                                            */
/*                                                                         */
/* RETURN VALUE:     Modified 'symbol' list and length of this list        */
/*                   (list is additional NULL terminated)                  */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Scans each 'symbol' of the actual line to be a STRING.*/
/*                   Each STRING is checked for be followed by optional    */
/*                   WHITESPACE and another STRING. 'symbols' maching this */
/*                   search pattern are tried to be merged by using the    */
/*                   'merge_str' function.                                 */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  If parameter is NULL, this is interpreted as an       */
/*                   empty list (same effect when first element is NULL).  */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
size_t merge_1line S1(symbol *,line)
{
  size_t i=0;                          /* loop variable                    */

  if(!line) return(0);                 /* return 0 on empty (undefined)    */
				       /* line -> contains no STRINGs      */

  while(line[i])                       /* scan line for STRINGs            */
  {
    if(line[i]->type==STRING)          /* STRING found ?                   */
    {
      size_t j;

      j=i+1;
      while(line[j]==WHITESP) j++;     /* skip optional WHITESPACE         */
      if(line[j])
      if(line[j]->type==STRING)        /* followed by another STRING ?     */
      {
	symbol s;

	if( (s=merge_str(line[i],line[j])) ) /* try to merge STRINGs...     */
	{
	  size_t k;

	  line[k=i]=s;                      /* ...and modify 'line' by     */
	  while( (line[++i]=line[++j]) );   /* removing the old 'symbols'  */
	  i=k-1;                            /* and adding the new 'symbol' */
	}                                   /* builded by 'merge_str'      */
      }
    }
    i++;                               /* continue scan-loop with the next */
  }                                    /* symbol (or new inserted 'symbol')*/

  do                                   /* cut off WHITESPACE at the end of */
  {                                    /* 'line' and count number of       */
    line[i]=NULL;                      /* 'symbols' on the remaining 'line'*/
    if(!i)break;
    i--;
  }
  while(line[i]==WHITESP);

  if(line[i]) i++;                     /* spec. handling for empty 'line'  */
  return(i);                           /* (processing "i++" is default)    */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         mrgstr_outp                                           */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         Produces output for a line (line_t type)              */
/*                                                                         */
/* PARAMETERS:       1) Pointer to a line (type line_t) or NULL            */
/*                   2) Flag: 0 suppresses call to 'output_line'           */
/*                   3) Flag: 0 suppresses call to 'output_sym'            */
/*                   4) Output mode for source line information            */
/*                      (just in effect if 1st parameter is NULL)          */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'output_line'/'output_sym'. Source line infor-  */
/*                   mation output for succeeding lines is suppressed for  */
/*                   output mode 1                                         */
/*                   There for this function needs static data, which is   */
/*                   initialized on calls with 1st parameter==NULL         */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  First call to this function must be a initialization  */
/*                   call (1st parameter==NULL, 4th parameter selects      */
/*                   output mode for further calls)                        */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void mrgstr_outp S4(line_t *,line,char,enlo,char,enso,char,lnos)
{
  static char *nxtname,                /* name and line number expected for*/
	      *nxtline;                /* the next call                    */
  static char omode;                   /* output mode for source line info.*/

  if(!line)                            /* initialization call ?            */
  {
    omode=lnos;                        /* set 'omode'                      */
    nxtname=nxtline=NULL;
    return;
  }
  if(omode==0) enlo=0;
  if(omode==1 && nxtline)              /* look if the line (number and     */
    if(!strcmp (nxtline,line->linen)   /* name) follows directly the last  */
      && !strcomp(nxtname,line->fname))/* line, which produced the contents*/
	enlo=0; 		       /* of 'nxtname' and 'nxtline'       */
				       /* in this case: suppress generating*/
				       /* source line inform. if 'omode'   */
				       /* was 1                            */
  free(nxtname);                       /* set 'nxtname'/'nxtline' for the  */
  nxtname=strdup(line->fname);         /* next line processed by this      */
  free(nxtline);                       /* function                         */
  nxtline=strdup(line->nlnno);
  if(enlo && line->linen)              /* output source line information   */
    output_line(line->outf,line->fname,line->linen,omode==1);
  if(enso)                             /* output the 'symbols' of a line   */
    output_sym (line->outf,line->cont);

  free_line(line);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         merge_strings                                         */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Merges all STRINGs on a stream of output lines        */
/*                   generated by the parser. And sends the remaining      */
/*                   lines (NULL terminated lists of 'symbols') to the     */
/*                   'output' module.                                      */
/*                   'merge_strings' replaces additional double empty      */
/*                   lines to single empty lines (as done before by the    */
/*                   lexical analysis, but processing of the parser can    */
/*                   generate new double empty lines!)                     */
/*                                                                         */
/* PARAMETERS:       1) 'stream'to use for output                          */
/*                   2) List of 'symbols' representing one line of the     */
/*                      (modified) source text as generated by the parser  */
/*                      for output (will be set 'free' by this function!!) */
/*                   3) Switch: -1 = normal operation, any value>=0 will   */
/*                                   initialize the output mode for source */
/*                                   line information:                     */
/*                               0 = generate output without source line   */
/*                                   information                           */
/*                               1 = same with source line information for */
/*                                   each line on change of filename or    */
/*                                   gap on line number counting (using    */
/*                                   #line-format, see 'output_line')      */
/*                               2 = source line information for each line */
/*                                   (using "name line:"-format, see ('out-*/
/*                                   put_line')                            */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'merge_strings' knows 6 different states (0..5).      */
/*                   See external documentation for details.               */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'stream' should be opened for output via 'open_w'     */
/*                   function (module 'output'), 1st parameter == NULL     */
/*                   selects no output (not used for this project).        */
/*                   First call to this function must be an initialization */
/*                   call. Don't use the un-initialized function!          */
/*                   Don't forget to call this function with a list        */
/*                   containing just the ENDOFFILE 'symbol' at the end of  */
/*                   all other actions of the program to flush the inter-  */
/*                   nal buffer!                                           */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void merge_strings S3(stream *,outf,symbol *,line,int,lnos)
{
  static int    state;                 /* actual state (-5..0..5)          */
				       /* negative states are used to exit */
				       /* this function with the           */
				       /* corresponding positiv state      */

  static line_t lempt,                 /* Last empty line     (buffered)   */
		lused,                 /* Last not empty line (buffered)   */
		actln;                 /* Actual line (parameter)          */

# define last_l                        /* last_l:                          */\
  (actln.cont[0]==ENDOFFILE)           /*   checks if actual line is the   */
				       /*   last line                      */

  if(lnos>=0)                          /* preform initialization           */
  {                                    /* (if selected) and return         */
    state=0;
    mrgstr_outp(NULL,0,0,lnos);
    return;
  }

  if(!outf)                            /* ignore calls with 1st parameter  */
  {                                    /* is NULL                          */
    free(line);
    return;
  }

  actln.len  =merge_1line(line);       /* set 'actln' for the actual line  */
  actln.cont =line;                    /* which was given as 2nd parameter */
  actln.outf =outf;
  actln.fname=strdup(get_name());
  actln.linen=strdup(get_line(0));
  actln.nlnno=strdup(get_line(1));

  do                                   /* loop (continued til 'state' is   */
				       /* negative!)                       */
  switch(state)
  {
    case 0 :                           /* goto 'state' 1 if first line is  */
      if(actln.len) state=1;           /* not empty, else: ignore first    */
	       else free_line(&actln); /* line and stay on 'state' 0       */
      break;
    case 1 :
      if(actln.cont[actln.len-1]->type==STRING)
      {
	lused=actln;                   /* buffer actual line if last       */
				       /* 'symbol' may be concatenateable  */
				       /* to a possibly following STRING   */
	state= -2;                     /* and exit function with 'state' 2 */
      }
      else                             /* else: generate output for the    */
      {                                /* actual line                      */
	mrgstr_outp(&actln,!last_l,1,0);
	state= -3;                     /* and exit function with 'state' 3 */
      }
      break;
    case 4 :
    case 2 :
      if(actln.len)                    /* actual line is not empty:        */
      {
	if(actln.cont[0]->type==STRING)/* actual line concatenateable with */
	{                              /* buffered line ('lused') ?        */
	  symbol s;

	  if( (s=merge_str(lused.cont[lused.len-1],actln.cont[0])) )
	  {
	    size_t i,n;                /* YES: perform concatenation       */
				       /*      (if possible)               */
	    lused.cont[lused.len-1]=s;
	    actln.len-=(n=(actln.cont[1]==WHITESP)?2:1);
	    for(i=0;i<=actln.len;i++)
	      actln.cont[i]=actln.cont[i+n];
	    if(state==4)               /*      and 'free' buffered empty   */
	      free_line(&lempt)        /*      if 'state' is 4             */
	    state=2;                   /*      goto 'state' 2 (check if new*/
	    break;                     /*      actual line is empty and    */
	  }                            /*      must be buffered)           */
	}
	mrgstr_outp(&lused,1,1,0);     /* NO : generate output for the un- */
				       /*      changed buffered line and   */
	if(state==4)		       /*      the succeeding buffered     */
	  mrgstr_outp(&lempt,!last_l,  /*      empty line (on 'state' 4)   */
		      !last_l,0);      /*      goto 'state' 1 (check if    */
	state=1;                       /*      unchanged actual line may be*/
	break;                         /*      concatenateable to the      */
      }		                       /*      succeeding lines)           */
      else                             /* actual line is empty:            */
      {
	if(state==4) free_line(&actln) /* 'free' actual line if 'state' was*/
		else lempt=actln;      /* already 4 (double empty line)    */
	state= -4;                     /* else buffer empty line and exit  */
      }                                /* with 'state' 4                   */
      break;
    case 3 :
      if(actln.len)                    /* actual line is not empy:         */
       state=1;                        /* goto 'state' 1                   */
      else
      {
	lempt=actln;                   /* else: buffer actual empty line   */
	state= -5;                     /* and exit with 'state' 5          */
      }
      break;
    case 5 :
      if(actln.len)                    /* actual line is not empty:        */
      {                                /* generate output for buffered     */
	mrgstr_outp(&lempt,!last_l,    /* empty line and goto 'state' 1    */
		    !last_l,0);        /* (further checks on the actual    */
	state=1;                       /* line)                            */
      }
      else                             /* else: skip doubel empty lines   */
      {                                /* (1 empty line was already       */
	free_line(&actln)              /* buffered)                       */
	state= -5;                     /* exit with 'state' 5             */
      }
  }
  while(state>0);

  state=(state<0)?-state:state;         /* leave 'merge_strings' with      */
}				        /* positive 'state' (static)       */
