/***************************************************************************/
/*                                                                         */
/* IDENTITY:         output                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  02.04.91                */
/*                   CODE             Schlotfeldt  30.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  23.05.91  10:32         */
/*                                                                         */
/* FUNCTION:         Should be used instead of stdio for opening 'streams' */
/*                   for output and writing characters to those 'streams'. */
/*                   Offers a more comfortable function handling objects   */
/*                   of the data type 'symbol' for output.                 */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'open_w', 'output_sym',             */
/*                       	       'output_text', 'output_line'        */
/*                                                                         */
/***************************************************************************/
#include "output.h"

#include "myalloc.h"
#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         open_w                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Creates a 'stream' object, which can be used as       */
/*                   parameter to the remaining functions of this module.  */
/*                   The 'stream' specifies a named FILE in the local      */
/*                   directory to write characters to.                     */
/*                                                                         */
/* PARAMETERS:       Name of FILE to create a 'stream' for or NULL for     */
/*                   building a 'stream' object for stdout                 */
/*                                                                         */
/* RETURN VALUE:     initialized 'stream' object as defined in 'types.h'   */
/*                   or NULL (error)                                       */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses 'fopen(...,'w')' for opening the named FILE or   */
/*                   builds a 'stream' object for stdout, if parameter is  */
/*                   NULL.                                                 */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: returns NULL on error ('fopen()' failed)              */
/*                                                                         */
/***************************************************************************/
stream *open_w S1(char *,s)
{
  stream *f;

  f=(stream *)malloc(sizeof(stream));  /* allocate new 'stream' object     */
  if(s)
  {                                    /* assign open FILE ...             */
    if( (f->fp=fopen(s,"w")) )
      f->name=strdup(s);
    else
    {                                       /* return NULL if 'fopen()'    */
      free(f);                              /* failed                      */
      f=NULL;
    }
  }
  else
  {                                    /* ... or assign stdout if name was */
    f->fp=stdout;                      /* NULL                             */
    f->name=NULL;
  }
  return(f);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         put_char                                              */
/*                                                                         */
/* TYPE:             Local macro                                           */
/*                                                                         */
/* FUNCTION:         Writes one character to a 'stream' or to stderr.      */
/*                   'stream' is automatically closed on output of EOF.    */
/*                                                                         */
/* PARAMETERS:       'stream' to use for output or NULL for stderr output  */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Checks wether output is to be done to a 'stream' or   */
/*                   direct to stderr (parameter is NULL).                 */
/*                   If 'stream' output is selected 'put_char' checks the  */
/*                   if the 'stream' is open. Output of EOF to an open     */
/*                   'stream' will close it.                               */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Calls to 'put_char' with a closed 'stream' will be    */
/*                   ignored (no action).                                  */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' if library function 'putc' returned    */
/*                   EOF (error, eg.: disk full)                           */
/*                                                                         */
/***************************************************************************/
#define put_char(c,f)                  \
 {                                     \
   if(f)                               /* 'stream' output ?                */\
   {                                   \
     if(f->fp)                              /* 'stream' open ?             */\
     {                                 \
       if(c==(char)EOF)                           /* close 'stream' if first     */\
       {                                    /* parameter is EOF            */\
	 fclose(f->fp);                \
	 f->fp=NULL;                   \
       }                               \
       else                                 /* any other characters are    */\
	 if(putc(c,f->fp)==EOF)             /* written to the connected    */\
	   errmsg(0,42,NULL,NULL);          /* FILE                        */\
     }                                 \
   }                                   \
   else putc(c,stderr);                /* output character on stderr if no */\
 }                                     /* 'stream' was selected            */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         output_sym                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Writes a list of 'symbols' (their text) to a 'stream' */
/*                   or to stderr.                                         */
/*                   Output is terminated by a newline character.          */
/*                                                                         */
/* PARAMETERS:       1) 'stream' to use for output or NULL for stderr      */
/*                      output                                             */
/*                   2) List of 'symbols' (NULL terminated) or NULL for    */
/*                      no output at all (except '\n' character)           */
/*                                                                         */
/* RETURN VALUE:     modified 'stream' (perhaps marked as closed)          */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Contains two loops: The first one processes the NULL- */
/*                   terminated 'symbol' list. The second processes the    */
/*                   single characters of the corresponding text of the    */
/*                   actual symbol. Some characters are inserted special   */
/*                   characters (see 'fetch' module), which need trans-    */
/*                   lation (done by a 'switch').                          */
/*                   The output of the characters is done by 'put_char'.   */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'stream' should be previosuly opened by 'open_w'      */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void output_sym S2(stream *,str,symbol *,sp)
{
  char *p;

  if(sp)                               /* check if any 'symbols' to be     */
				       /* processed by the following loop: */
  while(*sp)
  {
    p=(*sp++)->text;                   /* translate each character (non    */
    do switch(*p)                      /* special characters are left un-  */
       {                               /* changed -> default case)         */
	 case ESC    : put_char('\\',str); break;
	 case HASHM  : put_char('#' ,str); break;
	 case QUOTES : put_char('\\',str);
		       put_char('\"',str); break;
	 case APOSTR : put_char('\\',str);
		       put_char('\'',str); break;
	 case '\\'   : put_char('\\',str);
		       put_char('\\',str); break;
	 default     : put_char(*p  ,str);
       }
    while(*(++p));
  }
  put_char('\n',str);                  /* terminate output with newline    */
}                                      /* character                        */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         output_text                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Writes a text (string of characters) to a 'stream' or */
/*                   to stderr.                                            */
/*                                                                         */
/* PARAMETERS:       1) 'stream' to use for output or NULL for stderr      */
/*                      output                                             */
/*                   2) output text                                        */
/*                                                                         */
/* RETURN VALUE:     modified 'stream' (perhaps marked as closed)          */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Output is done by repeatedly calls to 'put_char'      */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'stream' should be previosuly opened by 'open_w'      */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void output_text S2(stream *,str,char *,s)
{
  while(*s)
  {
    put_char(*s,str);
    s++;
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         output_line                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Writes the "source line information" to a 'stream' or */
/*                   to stderr.                                            */
/*                   Format is: "filename line_number: "                   */
/*                          or  "line_number: "               for format 0 */
/*                          or  "#line line_number quoted_filename \n"     */
/*                          or  "#line line_number \n"        for format 1 */
/*                                                                         */
/* PARAMETERS:       1) 'stream' to use for output or NULL for stderr      */
/*                      output                                             */
/*                   2) Filename or NULL                                   */
/*                   3) Line number as string (as returned by function     */
/*                      'get_line' in module 'strinfo')                    */
/*                   4) Flag to select output format 0 or 1                */
/*                                                                         */
/* RETURN VALUE:     modified 'stream' (logical reasons might make it      */
/*                   impossible that the 'stream' is closed while this     */
/*                   operation)                                            */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Output is done by calls to 'output_text'.             */
/*                   Writing the filename is suppressed if 2nd parameter   */
/*                   is NULL.                                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'stream' should be previosuly opened by 'open_w'.     */
/*                   The second parameter mustn't be NULL!                 */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void output_line S4(stream *,str,char *,name,char *,line,char,form)
{
  if(form)                             /* output line number string for    */
  {                                    /* format 1                         */
    output_text(str,"#line ");
    output_text(str,line);
  }
  if(name)                             /* output name if parameter isn't   */
  {                                    /* NULL                             */
    if(form)
    output_text(str," \"");            /* put in quotes if format 1        */
    output_text(str,name);
    output_text(str,form?"\"":" ");
  }
  if(!form)                            /* output line number string for    */
  output_text(str,line);               /* format 0                         */

  output_text(str,form?"\n":": ");
}
