/***************************************************************************/
/*                                                                         */
/* IDENTITY:         srcstack                                              */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  16.03.91                */
/*                   CODE             Schlotfeldt  12.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  18:32         */
/*                                                                         */
/* FUNCTION:         Manages buffered input from one or more 'streams'     */
/*                   and 'macros'. Input sources (macro replace text       */
/*                   and/or input from 'streams') are handled on a stack.  */
/*                   Each 'push_xxx' adds a new element on the top of the  */
/*                   stack. Elements are poped when the last character of  */
/*                   their text/'stream' was read.                         */
/*                                                                         */
/* ------------------------------------------------------------------------

   %W% %D%

   Change History:
   ===============
   Mar-22-92:	'yytext' renamed to 'src_yytext' to avoid naming clashes with
   .		'lex' generated modules

   ------------------------------------------------------------------------*/
/*                                                                         */
/*                                                                         */
/* DATA/TABLES:      Contains complex static data (hidden in the not       */
/*                   global used function 'src_stack_op')                  */
/*                                                                         */
/* PROCEDURES:       for external use: 'init_srcstack', 'curr_stream',     */
/*                       	       'push_macro', 'push_stream',        */
/*		                       'input', 'unput', 'unput_all',      */
/*		                       'reduce', 'src_yytext'              */
/*		     for local use   : 'src_stack_op'                      */
/*                                                                         */
/***************************************************************************/
#include "srcstack.h"

#include "fetch.h"

#include "errmsg.h"
#include "stringu.h"
#include "myalloc.h"

enum cmd_t                             /* local type for calls to          */
     { input_,unput_all_,unput_,       /* 'src_stack_op' (function select) */
       yytext_,reduce_,push_stream_,
       push_macro_,curr_stream_,
       init_                        };

typedef union {                        /* local type for calls to          */
				       /* 'src_stack_op' (union of para-   */
				       /* meters for different functions): */
		char           c;      /*   input                          */
				       /*   unput_all                      */
		size_t         s;      /*   unput                          */
				       /*   reduce                         */
		stream        *f;      /*   push_stream, curr_stream       */
		char          *n;      /*   push_macro, src_yytext         */
	      }
	      src_stack_op_par;

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         src_stack_op                                          */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         'init_' function:                                     */
/*                     Initializes the local static variables and the      */
/*                     moduls 'fetch' and 'input' (via 'fetch')            */
/*                   'input_' function:                                    */
/*                     Get one character from the top 'stream' or 'macro'  */
/*                     replace text of the internal stack.                 */
/*                     The return value is stored in the 'c' part of the   */
/*                     parameter union.                                    */
/*                   'unput_' function:                                    */
/*                     Needs one parameter ('s' part of the par.union).    */
/*                     Unputs the specified number of characters. To the   */
/*                     'streams' / 'macros' they where read from.          */
/*                   'yytext_' function:                                   */
/*                     Returns a dynamic allocated string ('n' part of the */
/*                     parameter union) containing all characters that have*/
/*                     been read via 'input_' since initialization or last */
/*                     'reduce_' call. (includes 'reduce_' function at the */
/*                     end of processing!)                                 */
/*                   'unput_all_' function:                                */
/*                     Same as 'unput_' except all characters (see 'yy-    */
/*                     text_' function) are taken back.                    */
/*                   'reduce_' function:                                   */
/*                     Reduces internal buffers and initializes further    */
/*                     calls to 'input_'/'unput_' functions                */
/*                   'push_stream_' function:                              */
/*                     Pushes a new object (created internally) on the     */
/*                     internal stack. The next call to 'input_' function  */
/*                     will return the first character read from the       */
/*                     spec. 'stream' ('f' part of the par.union).         */
/*                     The return value (error flag for recursive          */
/*                     #includes) is stored in the 'c' part of the param.  */
/*                     union.                                              */
/*                   'push_macro_' function:                               */
/*                     Same as 'push_stream_'. Instead of a 'stream' a     */
/*                     'macro' replace text is used to read from.          */
/*                     (recursive 'macro' replacement is not checked)      */
/*                   'curr_stream_' function:                              */
/*                     Returns the highest entry of a 'stream' on the      */
/*                     stack (uses 'f' part of the par.union) or NULL, if  */
/*                     no 'stream' was found.                              */
/*                                                                         */
/* PARAMETERS:       1) Command to specify one of the 9 functions          */
/*                   2) Parameter (as union) for the specified function    */
/*                                                                         */
/* RETURN VALUE:     Modified parameter union                              */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/***************************************************************************/
/***************************************************************************/
/*                                                                         */
/* IDENTITY:         src_stack_op (continued)                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'init_' should be the first sub-function call.        */
/*                   'n' parameter to 'macro' replace text must be dynamic */
/*                   allocated.                                            */
/*                   Use 'push_xxx' functions only after initialization or */
/*                   'reduce_' call (this problem never occurs if 'lexi'   */
/*                   module is used to serve 'input_','unput_','unput_all_'*/
/*                   and 'reduce_' funtions!).                             */
/*                   Pushed 'streams' should be created by 'open_r'.       */
/*                                                                         */
/* ERROR CONDITIONS: Calls to 'push_xxx' functions at the wrong time may   */
/*                   cause serious internal errors.                        */
/*                   Don't try to read more characters than the maximum    */
/*                   value for the machine dependend type 'size_t' divided */
/*                   by 'sizeof(char)'!                                    */
/*                                                                         */
/***************************************************************************/
void src_stack_op S2(enum cmd_t,cmd,src_stack_op_par *,p)
{
# define BLKSIZE 512                   /* minimum size for memory blocks   */
				       /* allocated by 'src_stack_op'      */
				       /* high values minimize calls to    */
				       /* 'realloc' and 'free'             */
				       /* should be a power of 2 !!!       */

  typedef struct sse                   /* stack element type:              */
  {
    struct sse                   *nxt, /* pointer to next(higher)/previous */
				 *prv; /* element or NULL for last/first   */
				       /* element                          */
    char                         *buf; /* buffer for input characters read */
				       /* from 'stream' or copied from     */
				       /* 'macro' replace text             */
    struct                             /* pointer to 'buf':                */
    { size_t r,                        /*   pos. to read from              */
	     w,                        /*   pos. to write to               */
	     f,                        /*   pos. of first valid character  */
	     l; }		 p;    /*   last allocated pos.+1          */
    stream                       *f;   /* 'stream' to read from (to 'w'    */
				       /* position) or NULL if 'macro'     */
  }
  src_stack_el;

  static src_stack_el *stack;          /* internal stack                   */
  static size_t       blen;            /* characters read since last call  */
				       /* to sub-function 'reduce_'        */
  auto   src_stack_el *temp;
  auto   char         c;               /* actual character (for 'input_')  */
  auto   char         *txt;            /* return value for 'yytext_'       */

  switch(cmd)
  {
    case init_  :                      /* initialize static data           */
	 stack=NULL;
	 blen=0;
	 init_fetch();                 /* intialize 'fetch' module         */
	 return;

    case yytext_:                      /* 'yytext_' function (includes     */
	 if(!blen)                     /* 'unput_all_' several loop counts */
	 {                             /* in 'input_' and 'reduce_' funct.)*/
	   p->n=NULL;
	   return;
	 }

    case unput_all_:                   /* unput all characters read since  */
	 if(!(p->s=blen)) return;      /* last 'reduce_' call              */

    case unput_ :                      /* unput spec. ('p->s') number of   */
	 while(stack)                  /* characters                       */
	 {
	   if(stack->p.r-stack->p.f>=p->s)  /* unput as many as possible   */
	   {                                /* characters to the top stack */
	     stack->p.r-=p->s;              /* element and return if this  */
	     break;                         /* operation solves the problem*/
	   }
	   else
	   {
	     p->s-=stack->p.r-stack->p.f;
	     stack->p.r=stack->p.f;
	     if(stack->nxt)                 /* else process next stack     */
	       stack=stack->nxt;            /* element with remaining no.of*/
	     else                           /* characters or return if top */
	       break;                       /* of stack is reached         */
	   }
	 }
	 if(cmd!=yytext_)
	 {
	   blen-=p->s;                      /* exit for 'unput_' / 'unput- */
	   return;                          /* all functions               */
	 }
	 p->s=blen;                    /* continued 'yytext_' function:    */
	 txt=(char *)
	     malloc((blen+1)*sizeof(char)); /* allocate memory space for   */
	 txt[blen]=0;                       /* return value                */
	 blen=0;

    case input_ :                      /* get next character               */
	 if(cmd==input_) p->s=0;
	 do
	 {
	   blen++;
	   c=(char)EOF;                      /* ...is EOF for empty stack        */
	   while(stack)
	   {
	      while(stack->p.r>=stack->p.w  /* 'fetch' character if end of */
		 && stack->f)               /* buffer is reached and a     */
					    /* 'stream' will produce new   */
	      {                             /* characters                  */
		if(stack->p.w>=stack->p.l)
		{
		  size_t new;

		  new=(stack->p.l+=BLKSIZE)*sizeof(char);
		  if(new<=0)
		  {
		    src_stack_op(unput_all_,p);
		    errmsg(1,37,NULL,NULL);
		  }
		  stack->buf=(char *)realloc(stack->buf,new);
		}
		stack->buf[stack->p.w++]=fetch(stack->f);
	      }
	      c=stack->buf[stack->p.r];     /* store character as ret.value*/
	      if(c && (c!=(char)EOF || !stack->f))
	      {
		stack->p.r++;               /* increment 'r' pos. if not   */
		break;                      /* last character of 'stream'/ */
					    /* replace text is reached and */
					    /* "return"                    */
	      }
	      c=(char)EOF;
	      if(stack->prv)
		stack=stack->prv;           /* try to continue with prev.  */
	      else 			    /* stack element if EOF or NUL */
		break;			    /* where read                  */
					    /* "return" EOF if bottom of   */
	   }			 	    /* stack is reached            */
	   if(p->s)
	     txt[blen-1]=c;                 /* store character for 'yy-    */
	 }                                  /* text_' call and read further*/
	 while(blen<p->s);                  /* characters, loop ends im-   */
	 if(cmd==input_)                    /* mediatly for 'input_' call  */
	 {                                  /* (p->s was set to 0)         */
	   p->c=c;
	   return;                          /* exit for 'input_' function  */
	 }
	 p->n=txt;                     /* continued 'yytext_' function:    */

    case reduce_:                      /* reduce internal buffers and      */
				       /* initialize further calls to      */
				       /* 'input_'/'unput_' functions      */
	 if( (temp=stack) )
	 {
	   src_stack_el *tmp;          /* remove stack elements above the  */
				       /* actual stack pointer             */
	   tmp=temp->nxt;
	   while( (temp=tmp) )
	   {
	     if(temp->f)
	       free(temp->f->name);    /* remove complete 'stream' object  */
	     free(temp->f);            /* including allocated string for   */
				       /* his name                         */
	     free(temp->buf);          /* remove buffer (not longer in use)*/
	     tmp=temp->nxt;            /* remove stack element its self and*/
	     free(temp);               /* process next element above       */
	   }
	 }
	 else
	   return;                     /* return if stack is empty         */
	 stack->p.f=stack->p.r;        /* reduce element pointed to by     */
				       /* actual stack pointer             */
	 if(stack->p.f>=BLKSIZE        /* realloc buffer if too much       */
		&& stack->f)           /* garbage and element belongs to a */
	 {                             /* 'stream'                         */
           size_t i;

	   for(i=stack->p.f;i<stack->p.l;i++)
	     stack->buf[i-stack->p.f]=stack->buf[i];
	   stack->buf=(char *)realloc(stack->buf,stack->p.l-stack->p.f);
	   stack->p.l-=stack->p.f;
	   stack->p.w-=stack->p.f;
	   stack->p.f =stack->p.r=0;
	 }
	 stack->nxt=NULL;              /* cut off higher stack elements    */
				       /* previous removed                 */
	 blen=0;
	 return;

    case push_stream_:                 /* 'push_stream' function           */
	 temp=stack;                        /* check for #include file re- */
	 while(temp)                        /* cursion                     */
	 {
	   if(temp->f)
	   {                                /* by comparing all file names */
	     if(!strcomp(temp->f->name,     /* on the 'stack' with the new */
			 p->f->name   ))    /* one                         */
	     { p->c=1; return; }            /* return error ('p->c'=1)     */
	   }
	   temp=temp->prv;
	 }

    case push_macro_:                  /* 'push_macro' and continued 'push-*/
				       /* stream' function                 */
	 if(blen)
	   errmsg(0,27,"src_stack_op()",NULL); /* internal error */
	 if(stack)                     /* allocate new element on the old  */
	 {                             /* stack                            */
	   stack->nxt=(src_stack_el *)malloc(sizeof(src_stack_el));
	   stack->nxt->prv=stack;
	   stack=stack->nxt;
	 }
	 else
	 {                             /* ...or create new stack           */
	   stack     =(src_stack_el *)malloc(sizeof(src_stack_el));
	   stack     ->prv=NULL;
	 }
	 stack->nxt =NULL;             /* initialize new stack element     */
	 stack->p.r =stack->p.f= 0;
	 if(cmd==push_stream_)         /* for 'push_stream_'...            */
	 {
	   stack->f   =p->f;
	   stack->p.w =stack->p.l=0;
	   stack->buf =NULL;
	 }
	 else                          /* for 'push_macro_'...             */
	 {
	   stack->f   =NULL;
	   stack->p.w =stack->p.l=strlen(p->n)+1;
	   stack->buf =p->n;
	 }
	 p->c=0;                       /* return success for 'push_stream' */
	 return;                       /* functions                        */

    case curr_stream_:                 /* search and return the 'stream'   */
	 p->f=NULL;                    /* used in the highest stack pos.   */
	 temp=stack;
	 while(temp && !temp->f) temp=temp->prv;
	 if(temp)
	   p->f=temp->f;               /* return 'stream' or NULL (for no  */
	 return;                       /* 'stream' found                   */
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         input                                                 */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'input_'              */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Character read from the actual 'stream' / 'macro'     */
/*                   replace text                                          */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', return value is stored in       */
/*                   the 'c' part of the modified parameter union.         */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                   Don't try to read more characters than countable in   */
/*                   a variable of the machine dependend type 'size_t'!    */
/*                   Use frequently calls to 'reduce' function!            */
/*                                                                         */
/* ERROR CONDITIONS: Offending against 1st restriction will cause chaos!!  */
/*                                                                         */
/***************************************************************************/
char input S0(void)
{
  src_stack_op_par p;

  src_stack_op(input_,&p);
  return(p.c);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         unput_all                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'unput_all_'          */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', parameter union is just a       */
/*                   dummy for calling conventions to 'src_stack_op'       */
/*                   This dummy shouldn't be set to NULL, cause 'src-      */
/*                   stack_op' will use this union to store temporary      */
/*                   data.                                                 */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void unput_all S0(void)
{
  src_stack_op_par p;

  src_stack_op(unput_all_,&p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         unput                                                 */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'unput_'              */
/*                                                                         */
/* PARAMETERS:       Number of characters to write back to their source    */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', parameter is stored in the      */
/*                   's' part of the parameter union                       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void unput S1(size_t,s)
{
  src_stack_op_par p;

  p.s=s;
  src_stack_op(unput_,&p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         src_yytext                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'yytext_'             */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Dynamic allocated string                              */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', return value is stored in the   */
/*                   'n' part of the modified parameter union              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char *src_yytext S0(void)
{
  src_stack_op_par p;

  src_stack_op(yytext_,&p);
  return(p.n);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         reduce                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'reduce_'             */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op'                                  */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void reduce S0(void)
{
  src_stack_op(reduce_,NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         push_stream                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'push_stream_'        */
/*                                                                         */
/* PARAMETERS:       'stream' to push as element on the internal stack     */
/*                   of 'src_stack_op'. This 'stream' will be used for     */
/*                   further calls to 'input' til EOF is reached or        */
/*                   'stream' processing is interrupted by another call    */
/*                   to 'push_stream' or 'push_macro'.                     */
/*                                                                         */
/* RETURN VALUE:     Error flag for recursive #include files               */
/*                   (0=ok, 1=error)                                       */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', parameter is stored in the      */
/*                   'f' part of the parameter union                       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                                                                         */
/* ERROR CONDITIONS: Use this function only after initialization or calls  */
/*                   to 'reduce', otherwise 'src_stack_op' will abort with */
/*                   an internal error.                                    */
/*                                                                         */
/***************************************************************************/
int push_stream S1(stream *,f)
{
  src_stack_op_par p;

  p.f=f;
  src_stack_op(push_stream_,&p);
  return(p.c);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         push_macro                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'push_macro_'         */
/*                                                                         */
/* PARAMETERS:       Expanded replace text of a 'macro' which is to be     */
/*                   pushed as element on the internal stack of            */
/*                   'src_stack_op'.                                       */
/*                   The replace text will be used for further calls to    */
/*                   'input' til end of text (NUL) is reached or 'macro'   */
/*                   processing is interrupted by another call to          */
/*                   'push_stream' or 'push_macro'.                        */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', parameters are stored in the    */
/*                   'm' part of the parameter union                       */
/*                   'push_macro' called with parameter NULL will          */
/*                   immediately return with success (without any call to  */
/*                   'src_stack_op')                                       */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/*                   The replace text must be dynamic allocated, static    */
/*                   data may cause confusion if tried to set 'free'!      */
/*                                                                         */
/* ERROR CONDITIONS: Use this function only after initialization or calls  */
/*                   to 'reduce', otherwise 'src_stack_op' will abort with */
/*                   an internal error.                                    */
/*                                                                         */
/***************************************************************************/
void push_macro S1(char *,replc)
{
  src_stack_op_par p;

  if(!(p.n=replc)) return;
  src_stack_op(push_macro_,&p);
  return;
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         curr_stream                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See 'src_stack_op' sub-function 'curr_stream_'        */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Last 'stream' which was pushed by 'push_stream' on    */
/*                   the internal stack of 'src_stack_op' or NULL if no    */
/*                   'stream' was found                                    */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op', return value is stored in the   */
/*                   'f' part of the modified parameter union              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'src_stack' must be previously initialized     */
/*                   by 'init_srcstack'!                                   */
/* 								           */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
stream *curr_stream S0(void)
{
  src_stack_op_par p;

  src_stack_op(curr_stream_,&p);
  return(p.f);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_srcstack                                         */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initializes this module (including modules 'fetch'    */
/*                   and 'input')                                          */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'src_stack_op'                                  */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void init_srcstack S0(void)
{
  src_stack_op(init_,NULL);
}
