/***************************************************************************/
/*                                                                         */
/* IDENTITY:         strinfo                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  20.03.91                */
/*                   CODE             Schlotfeldt  16.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  18:39         */
/*                                                                         */
/* FUNCTION:         More comfortable interface to 'curr_stream' function  */
/*                   of module 'srcstack'                                  */
/*                   Extracts detailed informations from 'stream' object   */
/*                   delivered by 'curr_stream' and should be used instead */
/*                   of the direct call to 'curr_stream'.                  */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'get_name', 'get_line',             */
/*                       	       'get_date', 'get_time'              */
/*                                                                         */
/***************************************************************************/
#include "strinfo.h"

#include "srcstack.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_name                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets the name of the 'stream' returned by             */
/*                   'curr_stream' (call to this function included)        */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Name of the current name (may have been changed by    */
/*                   'ppLINE' function) of the last 'stream' that was      */
/*                   pushed on the 'srcstack' via 'push_stream' (see       */
/*                   module 'srcstack' for detailed information) or NULL   */
/*                   if no 'stream' was pushed.                            */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'curr_stream' and extracts name information     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char *get_name S0(void)
{
  stream *f;

  f=curr_stream();
  return(f?f->name:NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_line                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets the actual line number for the 'stream' returned */
/*                   by 'curr_stream' (call to this function included)     */
/*                                                                         */
/* PARAMETERS:       Offset to add to the returned line number             */
/*                                                                         */
/* RETURN VALUE:     Actual line number + offset (incremented on each 1st  */
/*                   character of a line or set by 'ppLINE' function) as   */
/*                   string.                                               */
/*                   'get_line' returns an empty string if no 'stream' was */
/*                   found to extract its actual line number (see 'curr-   */
/*                   stream')                                              */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'curr_stream' and extracts line number as       */
/*                   string. The size of the local static string 'line'    */
/*                   is guessed as 3 times the size of 'unsigned long' +1. */
/*                   (There can't be more than 3 decimal digits per byte!) */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Return value points to local static data! So 'strdup' */
/*                   the returned value if you don't want it to change its */
/*                   value with each call to this function!                */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char *get_line S1(int,offs)
{
  static char   line[3*sizeof(unsigned long)+1];
  auto   stream *f;

  f=curr_stream();
  if(f) sprintf(line,"%lu",f->line+offs); else line[0]=0;
  return(line);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_date                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets the date of opening the 'stream' returned        */
/*                   by 'curr_stream' (call to this function included)     */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Date when 'stream' was opened by 'open_r' (module     */
/*                   'input') in the form "MMM DD YYYY" where MMM is the   */
/*                   abbreviation for the Month DD the Day and YYYY the    */
/*                   Year.                                                 */
/*                   'get_date' returns an empty string if no 'stream' was */
/*                   found to extract its date of opening (see 'curr-      */
/*                   stream')                                              */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'curr_stream' and extracts the time of opening  */
/*                   this 'stream'. The time value is converted by 'ctime' */
/*                   to a string which is local static to 'ctime'.         */
/*                   The string (containing the time of day too) is merged */
/*                   to a minimum containing just the date information.    */
/*                   If no 'stream' was found variable 'd' is left un-     */
/*                   changed "1234", so 'd+4' will represent an empty      */
/*                   string (pointing to constant data)                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Return value points to local static data! So 'strdup' */
/*                   the returned value if you don't want it to change its */
/*                   value with each call to this function!                */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char *get_date S0(void)
{
  stream *f;
  char   *d="1234";                    /* dummy string initialization      */

  if( (f=curr_stream()) )
  {
    d=ctime(&f->time);                 /* convert time value to string     */
    d[24]=0;
    strcpy(d+11,d+20);
  }
  return(d+4);                         /* evaluates to "" if 'f' was NULL  */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_time                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets the time of opening the 'stream' returned        */
/*                   by 'curr_stream' (call to this function included)     */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     Time when 'stream' was opened by 'open_r' (module     */
/*                   'input') in the form "HH:MM:SS"                       */
/*                   'get_time' returns an empty string if no 'stream' was */
/*                   found to extract its time of opening (see 'curr-      */
/*                   stream')                                              */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'curr_stream' and extracts the time of opening  */
/*                   this 'stream'. The time value is converted by 'ctime' */
/*                   to a string which is local static to 'ctime'.         */
/*                   The string (containing the date too) is merged to a   */
/*                   minimum containing just the time information.         */
/*                   If no 'stream' was found variable 't' is left un-     */
/*                   changed "12345678901", so 't+11' will represent an    */
/*                   empty string (pointing to constant data)              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Return value points to local static data! So 'strdup' */
/*                   the returned value if you don't want it to change its */
/*                   value with each call to this function!                */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char *get_time S0(void)
{
  stream *f;
  char   *t="12345678901";             /* dummy string initialization      */

  if( (f=curr_stream()) )
  {
    t=ctime(&f->time);                 /* convert time value to string     */
    t[13]=t[16]=':';                   /* restore ':' (may be destroyed by */
    t[19]=0;                           /* 'get_date')                      */
  }
  return(t+11);                        /* evaluates to "" if 'f' was NULL  */
}
